#!/usr/bin/perl
# Vladimir Gabrielescu vgabriel@nbcs.rutgers.edu

# This cgi script maintains group changes and group change
# priviliges.
# 
# It is in fact a number of scripts combined into one by
# use of a switch board which uses hidden/Submit fields 
# to determine which action needs to be taken.
# It also uses a config file for most of it "intresting"
# variables. That config file (/usr/local/accounts/rats.conf)
# is documented internaly
#
# IMPORTANT: This program only uses /etc/group.gtmp (as defined in config)
# as a lock file. It uses both the empty file and a fcntl lock on it.
# however there are many other ways to lock group files and implementing
# all of them would be a nightmare. For example admintool, uses a fcntl lock
# on the group file and /etc/.group.lock, groupadd, uses a fcntl lock on
# /etc/gtmp, groupmod/usermod use no locking whatsoever. To make things
# even more fun makeclass uses /etc/group.gtmp and/or /etc/.gtmp.  We are
# more worried about makeclass as the sun tools are less likley to be used,
# so we'll use /etc/group.gtmp
#
# Things to note. 
# -- super users (users who can admin groups without being in them)
#    are defined in the config file. Their privs can only be removed
#    and added by hand. To do so they must by added to the super hash
#    in the config file and their privs added to the privs file (defined
#    in the config file)
# -- the privs file matches a user to the groups for which he/she
#    has priviliges. This script will edit entries unless they are of
#    super users.
# -- the "users" file is generated from the privs file and it's used
#    by the radius rule in the .htaccess file.
# -- extra groups ( groups created to account for entry length limits)
#    will have the format: <orig group>-<ext-token><digit>. The ext-token
#    is defined in the config file
# -- by default all the users will be authenticated using the radius domain
#    defined in the config file. users added manualy to the privs/users 
#    and superusers can use any domain the httpd server supports
# -- removing a user from a group will not remove the change privs for that
#    user for that group. That operation must be performed separately
# -- logging is done to a file due to perl syslog problems
# -- there is a limit on the length of a group entry around 1024chars. By 
#    default we set the limit to 900. Change it in config file.
# -- if $CONFIG::YPFILE exists then a button on the main screen will
#    make it avaible

use File::Copy;
use FileHandle;
use strict;
use CGI qw(:standard);
use vars qw(%States $page @groups $grname $gid $pass $members);
use vars qw($user %privs $auth_user $lock_fh $log $debug);
use vars qw($do $c_port $s_port $target_host);

require "/usr/local/accounts/lib/RatsClientLib.pm";
require "/usr/local/accounts/etc/rats_internal.conf";
require "/usr/local/accounts/lib/RatsClientStyle.pm";




# we define the log routine for all to see right here
$log = $CONFIG::COMMAND_HOOKS{'LOG'};
# same for debug
$debug = $CONFIG::COMMAND_HOOKS{'DEBUG'};

# remote call function
$do = \&RatsClientLib::send_remote_command;
$c_port=$CONFIG::CLIENT_PORT;
$s_port=$CONFIG::SERVER_PORT;
$target_host = $CONFIG::MASTER_HOST;

## the user gets set by httpd from the AUTH part
$auth_user=$ENV{REMOTE_USER};

## create a new query and set the header
my $q = new CGI;


print $q->header;
my $title = $CONFIG::VIGR_WEB_TITLE;
print $q->start_html(-title=>$title,
		     -style=>{'src'=>$RatsStyle::StyleSheetLocation},
		     -dtd=>'-//W3C//DTD HTML 4.0 Transitional//EN',
		    );
  
&RatsStyle::doc_frame_nomenu($title);
&RatsStyle::start_cgi();


## And if the user is not set then we quit
if ($auth_user eq "") {
    print "Unauthenticated. Bye.";
    &RatsStyle::end_cgi();
    &RatsStyle::copyright_center();
    &RatsStyle::doc_end();
    exit -1;
}

# this gets the current state of the script
$page = $q->param(".State") || $q->param("Reset") || "Default";


## This allows for the many in one script. 
%States = (
	   'Default' => \&choose_group,
	   'Display' => \&display_users,
	   'ListUsers' => \&list_users,
	   'AddUserScreen' => \&add_user_screen,
	   'AddUser' => \&add_user,
	   'RemoveUserScreen' => \&remove_user_screen,
	   'RemoveUser' => \&remove_user,
	   'ChangePrivs' => \&change_privs_screen,
	   'Change' => \&change_privs,
	   'PushChanges' => \&push_changes,
	   'Help' => \&display_help,
);


# we dispatch to the appropriate routine
if ($CONFIG::WEB_NO_SERVICE) {
    print &error($CONFIG::WEB_NO_SERVICE_MSG);
} else {
    if (exists $States{$page}) {
	$States{$page}->();
    } else {
	print "No such page";
    }
}
&RatsStyle::end_cgi();
&RatsStyle::copyright_center();
&RatsStyle::doc_end();
exit;

##### end of main

# this routine lists the users in a group
sub list_users {
    my $group = $q->param("Group");
    $group =~ s/\s+$//g;
    unless (&check_priv($group)) {
      &$log("$auth_user attempted to list users for $group. Unauthorized");
      print "You are not authorized to edit this group. Attempt logged.";
      return;
    }
    my ($gn, $gp, $gid, $mem, $time);
    ($gn, $gp, $gid, $mem, $time) = remote_getgrnam($group);
    print
	$q->h1('List of users for a group'),
	"The users in group \"$group\" are:",$q->p,
	"$mem",
	$q->hr;
    &choose_group;
}




# main front page which selects the group and the action
sub choose_group {
    my $temp = &get_privs($auth_user);

    unless ($temp) {
      print "You have no group editing privs.";
      return;
    }


    @groups = split(",", $temp);
    print 
	$q->h1('Choose a Task'),
	"Hello, $auth_user.",$q->p,$q->hr,
	$q->start_form,
	"What group do you wish to manage?\t",
	$q->popup_menu(-name=>'Group',
		   -values=>[@groups]), p,
	to_page("ListUsers"),
	to_page("AddUserScreen"),
	to_page("RemoveUserScreen"),
	to_page("ChangePrivs");
    # and we only offer the yp push option if the yp update file exits.
    if (-e $CONFIG::YPFILE) {
	print to_page("PushChanges");
    }
    print to_page("Help");
    print
	$q->end_form,
	$q->hr;
}


# this routine find out what groups a user is allowed to change
sub get_privs ($){
    my $current = $_[0];
    my ($temp_user, $privs, $final_privs, $super_user, @temp_array);
    $final_privs = "";
    if ( -f $CONFIG::PRIVS_FILE ) {
	open (ACCT, "<$CONFIG::PRIVS_FILE") || print "Failed to open privs file $CONFIG::PRIVS_FILE",p;
    } else {
	print &error("Failed to get privs file $CONFIG::PRIVS_FILE. Please contact the site admin.");
	return;
    }
    while (<ACCT>) {
	if (/^\#/) { # ignore comments
	    next;
	}
	($temp_user, $privs) = split(":");
	if ($temp_user eq $current) {
	    $privs =~ s/\s+//g;
	    $final_privs = $privs;
	}
    }
    # we merge in whatever super user privs
    $super_user = &is_super($current);
    if ($super_user ne "" && $CONFIG::GROUP_SUPER{$super_user} ne "") {
	if ($final_privs ne "") {
	    $final_privs = join (",", $final_privs, $CONFIG::GROUP_SUPER{$super_user});
	} else {
	    $final_privs = $CONFIG::GROUP_SUPER{$super_user};
	}
    }
    # now we clean it up a bit
    @temp_array = split(",", $final_privs);
    @temp_array = &sort_unique(@temp_array);
    $final_privs = join(",", @temp_array);

    # and we return it
    return $final_privs;
}

# this is the screen on which a user can be added to a group
sub add_user_screen {
    my $group = $q->param('Group');

    unless (&check_priv($group)) {
      &$log("$auth_user attempted to add users to $group. Unauthorized");
      print "You are not authorized to edit this group. Attempt logged.";
      return;
    }

    print
	$q->start_form,
	$q->hidden(-name=>'Group', -default=>$group),
	"What is the users you wish to add to group \"$group\"? ",
	"(comma separated) ",
	$q->textfield('User'),p,
	to_page("AddUser"),
	$q->end_form,
	$q->hr;
}

#verify that user is allowed to access group
sub check_priv() {
  my ($group) = @_;
  my $temp_priv = &get_privs($auth_user);
  my @temp_groups = split(",", $temp_priv);
  my $tg;
  foreach $tg (@temp_groups) {
    if ($group eq $tg) {return 1};
  }
  return 0;
}


# this is the screen on which users can be selected for group removal
sub remove_user_screen {
    my($group, $gn, $gp, $gid, $mem, $time, @users);
    $group = $q->param('Group');
    $group =~ s/\s+//g;
    unless (&check_priv($group)) {
      &$log("$auth_user attempted to remove users from $group. Unauthorized");
      print "You are not authorized to edit this group. Attempt logged.";
      return;
    }
    ($gn, $gp, $gid, $mem, $time) = remote_getgrnam($group);
    @users = split(",", $mem);
    print
	$q->start_form,
	$q->hidden(-name=>'Group', -default=>$group),
	"Select the users you wish to remove from group \"$group\" ",$q->p,
	$q->checkbox_group(-name=>'users',
			   -values=>\@users,
			   -columns=> 4), $q->p,
	to_page("RemoveUser"),
	$q->end_form,
	$q->hr;
}


# add a user to a group and log
sub add_user {
    my($group, $user, $result, $code, @data, $user_list, @users);
    $group = $q->param('Group');
    $user_list = $q->param('User');
    unless (&check_priv($group)) {
      &$log("$auth_user attempted to add $user_list from $group. Unauthorized");
      print "You are not authorized to edit this group. Attempt logged.";
      return;
    }
    chomp $user_list;
    if ($user_list eq "") {
	print "No user was given!",p;
	&add_user_screen;
    }

    @users = split(",", $user_list);
    foreach $user (@users) {
      if (&user_allowed($user)) {
	$result = &$do('EXUSER', $user, $target_host,$c_port);
	($code, @data) =  split(":", $result);
	if ($code  == $CONFIG::API{'R_EXUSER'}) {
	    $result = &$do('ADDTOGRP', $user, $group, $target_host,$c_port);
	    ($code, @data) =  split(":", $result);
	     if ($code  == $CONFIG::API{'R_ADDTOGRP'}) {
		&$log("$auth_user added $user to $group\n");
		print "User $user added to group $group", $q->p;
	    } else {
		&$log("$auth_user failed to add $user to $group\n");
		print "Failed add $user to group $group",$q->p;
		print "Please contact root\@$CONFIG::MASTER_HOST!",$q->p;
	    }

	} else {
	    &$log("$user is not a valid user\n");
	    print "User $user is not a valid user on this host",$q->p;
	}
      } else {
	&$log("$auth_user tried to add prohibted user $user to $group\n");
	print "$user may not be added to any groups using this tool",$q->p;
      }
    }
    print $q->hr;
    &choose_group;
}


# remove a user from a group and log
sub remove_user {
    my ($privs, @privs, $priv);
    my ($group, $user, $result, $code, @data);
    $group = $q->param('Group');
    unless (&check_priv($group)) {
      &$log("$auth_user attempted to remove users from $group. Unauthorized");
      print "You are not authorized to edit this group. Attempt logged.";
      return;
    }

    foreach $user ($q->param('users')) {
      if (&user_allowed($user)) {
        $result = &$do('REMFRGRP', $user,$group, $target_host,$c_port);
        ($code, @data) = split(":", $result);
        unless ($code == $CONFIG::API{'R_REMFRGRP'}) {
	    &$log("$auth_user failed to add $user to $group\n");
	    print "Failed remove $user from group $group",$q->p;
	    print "Please contact root\@$CONFIG::RADIUS_DOMAIN!",$q->p;
        } else {
	    &$log("$auth_user removed $user from $group\n");
	    print "User $user removed from group $group",$q->p;
	    # check if the user had any privs for that group and complain
	    $privs = &get_privs("$user\@$CONFIG::RADIUS_DOMAIN");
	    @privs = split(",", $privs);
	    foreach $priv (@privs) {
		if ($priv eq $group) {
		    print "Note that user $user still has privs for group $group.";
		    print "You should make sure that the privs are removed as ";
		    print "well if you so desire.";
		    print $q->p;
		    &$log("$user removed from $group but still has privs for $group\n");
		}
		
	    }
	  }
      } else {
	&$log("$auth_user tried to remove prohibted user $user from $group\n");
	print "$user may not be removed from any groups using this tool",$q->p;
      }
    }
    print $q->hr;
    &choose_group;

}


# this allows the maintainer to select and remove users who can 
# modify a group.
sub change_privs_screen {
    my($gn, $gp, $gid, $mem, $group, @users, $privs, @old, $temp_user);
    my(@privs, $temp_priv, $domain, $real_user, %users, $time);

    $group = $q->param("Group");

    $group =~ s/\s+//g;

    unless (&check_priv($group)) {
      &$log("$auth_user attempted to change privs for $group. Unauthorized");
      print "You are not authorized to edit this group. Attempt logged.";
      return;
    }


    ($gn, $gp, $gid, $mem, $time) = remote_getgrnam($group);    
    @users = split(",", $mem);
    foreach $temp_user (@users) {
	$users{$temp_user} = 1;
    }
    
    ## find the current prived users
    open (ACCT, "<$CONFIG::PRIVS_FILE") || print "Could not open privs file!\n";
    while (<ACCT>) {
	if (/^\#$/) {
	    next;
	}
	chomp;
	($temp_user, $privs) = split(":");
	($real_user, $domain) = split("\@", $temp_user);
	@privs = split(",", $privs);
	foreach $temp_priv (@privs) {
	    if ($temp_priv eq $group) {
		push @old, $real_user;
	    }
	}
    }
    
    ## find the Super users and list them as well
    foreach $temp_user (keys %CONFIG::GROUP_SUPER) {
	($real_user, $domain) = split("\@", $temp_user);
    	@privs = split(",", $CONFIG::GROUP_SUPER{$temp_user});
	foreach $temp_priv (@privs) {
	    if ($temp_priv eq $group) {
		push @old, $real_user;
	    }
	}
    }

    #combine the users in the group with those in the privs
    push @users, @old;
    @users = &sort_unique(@users);

    print
	"The current privliged users in this group are already selected. ",
	"Select and deselect users who can add users and ",
	"change user privs for group \"$group\":",p,
	$q->start_form,
	$q->hidden(-name=>'Group', -default=>$group),
	$q->checkbox_group(-name=>'users',
			   -values=>\@users,
			   -defaults=>\@old,
			   -columns=> 4), $q->p,
	to_page("Change"),$q->p,
	"NOTE: even if you remove the privileges of a 'Super User', ",
	"they will reapear next time you run this program. Only ",
	"system administrators can remove the privs of a 'Super User'.",
	$q->end_form,
	$q->hr;
}


# and this actualy changes the privs, regenerates
# privs and users files
sub change_privs {
    my($temp_user, %users, $privs, @old_privs, $priv, %new_users);
    my($new_entry, $group, @priv_users, %old_users, @users);
    my($array_pointer, @final, @temp_privs, $temp_priv, $found);
    my($super_name, @auth_array);
    # the group we are looking at. it comes from hidden field
    $group = $q->param('Group');
    chomp $group;
    unless (&check_priv($group)) {
      &$log("$auth_user attempted to change privs for $group. Unauthorized");
      print "You are not authorized to edit this group. Attempt logged.";
      return;
    }
    # this is powered by the checkbox from the previous page
    @users = $q->param('users');
    
    # build a hash for which the value is the name of the group
    foreach $user (@users) {
	# if the user is a super we only add the entry if he doesn't
	# already have it
	$super_name=&is_super($user);
	if ($super_name ne "") {
	    $found = 0;
	    @temp_privs = split(",",$CONFIG::GROUP_SUPER{$super_name});
	    foreach $temp_priv (@temp_privs) {
		if ($temp_priv eq $group) {
		    $found = 1;
		}
	    }
	    unless ($found) {
		$users{"$super_name"} = $group;
		push @final, $super_name;
	    }
	} else {
	  if ($CONFIG::RADIUS_DOMAIN) {
	    $users{"$user\@$CONFIG::RADIUS_DOMAIN"} = $group;
	  } else {
	    $users{"$user"} = $group;
	  }
	  push @final, $user;
	}
    }

    # now we lock the privs file so noone touches while we muck around
    # we give away the lock much later, when we are done with all the
    # manipulations. Note that we never really write into the privs
    # file directly, we just lock it.
    my $acct = FileHandle->new(); # get a filehandle so we can pass it 
    open ($acct, ">>$CONFIG::PRIVS_FILE") || print "Could not open privs file!\n";
    GetFileLock($acct);

    # note that the accounts file must exits otherwise 
    # you couldn't get this far.
    if ( -f $CONFIG::PRIVS_FILE ) {
	open (ACCT, "<$CONFIG::PRIVS_FILE") || print "Failed to open privs file";
    } else {
	print "Failed to get privs info. Please contact the site admin.";
	exit -1;
    }
    # and we build a hash out this existing entries
    while (<ACCT>){
	$temp_user = $privs = "";
	if (/^\#/) {
	    next;
	}
	chomp;
	($temp_user, $privs) = split(":");
	$old_users{$temp_user} = $privs;
    }
    close ACCT;


    
    ## now we merge the two hashes
    #  first we look at all the old users and add or remove 
    #  privs as requiered
    foreach $user (keys %old_users) {
	my (@temp_array, %temp_hash, $temp_priv);
	@temp_array=split(",", $old_users{$user});
	# add privs if need it
	if ($users{$user} ne "") {
	    push @temp_array, $users{$user};
	    @temp_array = &sort_unique(@temp_array);
	    $old_users{$user} = join(",", @temp_array);
	    delete $users{$user};                  # so we can ignore it later
	} else {       # we check if it had the privs and it should anylonger
	    foreach $temp_priv (@temp_array) {
		$temp_hash{$temp_priv} = 1;
	    }
	    # note that the SUPER users privs can be removed but the 
	    # privs in the config file will reapear
	    if ($temp_hash{$group}) {
		delete $temp_hash{$group};
		&$log("$auth_user removed privs for $user for group $group\n");
	    } 
	    @temp_array = keys %temp_hash;
	    $old_users{$user} = join(",", @temp_array);
	    # and we keep track of Supers to be nice to the user at the end
	    if ($CONFIG::GROUP_SUPER{$user} && $temp_hash{$group}) {
		push @final, $user;
	    }
	}
    }

    #  now we look at all the new users and we merge them in old_users
    #  the ones which were there before have been merged and deleted so
    #  all the ones left can be just added.

    foreach $user (keys %users) {
	$old_users{$user} = $users{$user};
	&$log("$auth_user added privs for $user for group $group\n");
    }
    
    # now old_users is really new_users :-)
    # so we first clean up junk entries
    foreach $user (keys %old_users) {
	$new_users{$user} = $old_users{$user} unless  ($old_users{$user} =~ /^\s*$/);
    }
    # and we save the data to a temp file
    open (TMP, ">$CONFIG::PRIVS_FILE_TEMP") || print "Could not open $CONFIG::PRIVS_FILE_TEMP!\n";
    foreach $user (keys %new_users) {
	print TMP "$user:$old_users{$user}\n" ;
    }
    close TMP;
    
    ## update the radius AUTH file
    #  all we really need there is one line entries for all the prived
    #  users. We combine it from the list of prived useres and the list
    #  of super users 
    open (TMP, ">$CONFIG::RAD_AUTH_FILE_TEMP") || print "Could not open tmp radius authfile!\n";
    @auth_array = keys %new_users;
    push @auth_array, keys %CONFIG::GROUP_SUPER;
    @auth_array = &sort_unique(@auth_array);
    foreach $temp_user (@auth_array) {
	print TMP "$temp_user\n";
    }
    close TMP;

    
    ## and now we move this to files to real names
    tmove ("$CONFIG::RAD_AUTH_FILE_TEMP", "$CONFIG::RAD_AUTH_FILE") || print "Copy failed\n";
    #  we need to chmod it so httpd can read it.
    chmod 0644, $CONFIG::RAD_AUTH_FILE;
    tmove ("$CONFIG::PRIVS_FILE_TEMP", "$CONFIG::PRIVS_FILE")|| print "Copy failed\n";
    GiveFileLock($acct);
    close $acct;
    
    # and say something nice to the user
    print 
	"Done. Now the following users have",
	" maintainace privileges for group \"$group\" :",p;
    @final = &sort_unique(@final);
    foreach $user (@final) {print $user,p;}
    print $q->hr;
    &choose_group;
}

# get exclusive file lock
sub GetFileLock ($) {
    use Fcntl ':flock';                        # import LOCK_*
    if (flock($_[0], LOCK_EX)) {               # wait for lock
        seek($_[0], 0, 2);                     # make sure you have the
    } else {                                   # begining of the file
        print "GetFileLock: Could not get file lock for $_[0]: $!",p;
    }
}
 

# drop file lock
sub GiveFileLock ($) {
    use Fcntl ':flock';        # import LOCK_*        
    flock($_[0], LOCK_UN);     # and unlock
}

# sort unique of a given array
sub sort_unique (@) {
    my %saw_hash;
    @saw_hash{@_} =();
    return sort keys %saw_hash;
}


# equivalant of getgrnam but it does the look up through RATS daemon
# It is sort of stupid that we return nulls if it fails
sub remote_getgrnam ($) {
    my $group = $_[0];
    my ($final_gn, $final_gp, $final_gid, $final_mem, $timestamp);
    my ($result, $code, @data);
    $result = &$do('GETGRNAM', $group, $target_host,$c_port);
    ($code, @data) = split(":", $result);
    if ($code == $CONFIG::API{'R_GETGRNAM'}) {
	($final_gn, $final_gp, $final_gid, $final_mem, $timestamp) = @data;
    } else {
	&$log("Failed to get list membership for $group\n");
    }
    return ($final_gn, $final_gp, $final_gid, $final_mem,$timestamp);
}


# this routine moves the file given by arg1 to arg2 
sub tmove ($$) {
    my ($source, $target) =@_;
    # first we try the system way
    return 1 if rename $source, $target;
    # ok so that didn't work . So we do a tcopy and unlink
    copy ($source, $target);
    unlink $source || print "Could not unlink $source: !$\n";
    return 1
}

# if the appropriate update file exists then run it and display
# the output
sub push_changes() {
    if (-e $CONFIG::YPFILE) {
	print "Found push script.",$q->p;
	open (YP, "$CONFIG::YPFILE|") || die "Could not open yp: $!\n";
	while (<YP>) {
	    print $_,$q->p;
	}
	print "Changes pushed.",$q->p;
	&$log("$auth_user pushed changes\n");
    } else {
	print "No push ability administratively configured.\n",$q->p;
    }
    print $q->hr;
    &choose_group;
}

# display help
sub display_help {
    print
	$CONFIG::WEB_GROUP_HELP,
	$q->hr;
    &choose_group;
}

# here we take a username as an arg, if username@anything exists in
# the super user hash then we return username@anything else ""
# only the first match is returned
sub is_super ($) {
    my ($user, $super_user, $real_user, $domain, $scrap);
    ($user,$scrap) = split ("\@",$_[0]);
    foreach $super_user (keys %CONFIG::GROUP_SUPER) {
	($real_user, $domain) = split ("\@", $super_user);
	if ($real_user eq $user) {
	    return $super_user;
	}
    }
    return "";
}


sub user_allowed() {
  my $user = $_[0];
  my $temp_user;
  foreach $temp_user (@CONFIG::GROUP_PROHIBTED_USERS) {
    if (lc($user) eq lc($temp_user)) {return 0;}
  }
  return 1;
}

# submit button with a hidden field.
# its only arg should be the value of the hidden state
sub to_page ($) {
    my $val = shift;
    return &RatsStyle::to_page($q,$val);
}
# format error message
sub error ($) {
  return &RatsStyle::error($_[0]);
}

# format info message
sub info ($) {
  return &RatsStyle::info($_[0]);
}

