#!/usr/bin/perl


package RatsLib;
require 5.001;

require "/usr/local/accounts/lib/RatsClientLib.pm";
require "/usr/local/accounts/etc/rats_internal.conf";


use Fcntl ':flock';                        # import LOCK_*
use Quota;
use Socket;
use Sys::Hostname;
use File::Basename;
use IO::Socket;
use File::Path;
use FileHandle;
use File::Copy;
use Carp;
require Exporter;
@ISA = qw(Exporter);



use strict;

# field names in local files this code can modify
my %passwd_name2num = (
		       'uname'    => 0,
		       'pass'     => 1,
		       'uid'	  => 2,
		       'gid'	  => 3,
		       'gcos'	  => 4,
		       'home'	  => 5,
		       'shell'	  => 6,
		       );

my %group_name2num = (
		       'gname'    => 0,
		       'pass'     => 1,
		       'gid'	  => 2,
		       'mem'	  => 3,
		       );

my %shadow_name2num = (
		       'uname'    => 0,
		       'pass'     => 1,
		       'lastchg'  => 2,
		       'mindays'  => 3,
		       'maxdays'  => 4,
		       'warn'	  => 5,
		       'inactive' => 6,
		       'expire'	  => 7,
		       'flag'	  => 8,
		       );

my %reserve_name2num = (
			'uname'    => 0,
			'uid'      => 1,
			);

my %reserveg_name2num = (
			 'uname'    => 0,
			 'gid'      => 1,
			 );


# we define the log routine for all to see right here
my $log = sub {
    &{$CONFIG::COMMAND_HOOKS{'LOG'}};
};
# same for debug
my $debug = sub {
    &{$CONFIG::COMMAND_HOOKS{'DEBUG'}};
};
# remote call function
my $do = sub {
    RatsClientLib::send_remote_command(@_);
};



# takes a quota setting arg list (uid:soft:hard:fs), figures
# out if the file system is local or remote and dispatches the call to
# the appropriate entity
sub default_set_quota ($$$$$) {
    my ($uid,$soft,$hard,$fs) = @_;

    &$debug("Got $uid $soft $hard $fs\n");

    # check $fs and untaint
    if ($fs =~ /^($CONFIG::PATH_REGEXP)$/) {
        $fs = $1;                     # $fs now untainted
    } else {
        &$log("Bad data in $fs\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }  

    my $new_fs = &RatsClientLib::determine_fs($fs);
    if ($new_fs eq "") {
	&$log("Can't determine fs for $fs\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'NFS'}";
    } else {
	$fs = $new_fs;
    }
    
    my $host = &get_hostname();
    my $target_host = $CONFIG::FS{$fs};
    if ($host eq $target_host) {
	&$debug("spawn local $uid $soft $hard $fs\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'LQSET'};
	return &$handler_sub($uid,$soft,$hard,$fs);
    } else {
	&$debug("spawn remote $uid $soft $hard $fs $target_host\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'RQSET'};
	return &$handler_sub($uid,$soft,$hard,$fs,$target_host,$CONFIG::CLIENT_PORT);
    }
}

# Dispatches quota call to appropriate host. It takes 
# uid:soft:hard:fs:host as args
# and return 1 for success, -1 for connection failure,  0 otherwise
#
# It only sends the arg list and the command name to the
# generic client sender
sub default_remote_set_quota ($$$$$$) { 
    return &RatsClientLib::send_remote_command("QSET",@_);
}

# set quotas on local FS. It takes uid:soft:hard:fs as args
# and return 1 for success and 0 for failure.
#
#
sub default_local_set_quota ($$$$$) {
    my ($uid,$soft,$hard,$fs) = @_;
    my $return;
    &$debug("Got $uid $soft $hard $fs\n");
    
    # figure out the device for the quota
    my $dev =  Quota::getqcarg($fs);

    # set the quota
    &Quota::setqlim($dev, $uid, $soft, $hard, 0, 0, 0);

    #check error status
    my $error = Quota::strerr();
    unless ($error eq "error \#0") {
	&$log("failed to set quota: $error\n");
	return "$CONFIG::API{'CODEDERR'}:$error";
    }
    &$debug("set quota for $uid on $fs succesful.\n");
    return "$CONFIG::API{'R_QSET'}:1";
}

# create a directory by figuring out the right file system
# and dispatching the request localy or remotly
sub default_mk_dir ($$$$) {
    my ($uid,$gid,$path,$perm) = @_;
    # figure our where the actual fs is 
    # or if it even exists
    &$debug("Got $uid $gid $path $perm\n");
    my $fs = &RatsClientLib::determine_fs($path);
    if ($fs eq "") {
	&$debug("Can't determine fs for $path\n");
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NFS'}";
    }
    &$debug("fs for $path is $fs\n");
    my $host = &get_hostname();
    my $target_host = $CONFIG::FS{$fs};
    if ($host eq $target_host) {
	&$debug("spawn local $uid $gid $path $perm\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'LMKDIR'};
	return &$handler_sub(@_);
    } else {
	&$debug("spawn remote $uid $gid $path $perm $target_host\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'RMKDIR'};
	return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);
    }
}

# create a local directory with the owner, permissions and group given
sub default_local_mk_dir ($$$$) {
    my ($uid, $gid, $path, $perm) = @_;
    # check path and untaint
    if ($path =~ /^([-\w.\/]+)$/) {
        $path = $1;                     # $path now untainted
    } else {
        &$log("Bad data in $path\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }
    # check perms and untaint
    if ($perm =~ /^(\d+)$/) {
        $perm = $1;                     # $perm now untainted
    } else {
        &$log("Bad data in $perm\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }
    # check uid and untaint
    if ($uid =~ /^(\d+)$/) {
        $uid = $1;                     # $uid now untainted
    } else {
        &$log("Bad data in $uid\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }
    # check gid and untaint
    if ($gid =~ /^(\d+)$/) {
        $gid = $1;                     # $gid now untainted
    } else {
        &$log("Bad data in $gid\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }

    # create the directory
    unless ( -e $path) {
	mkpath ($path, 0, oct $perm) || 
	    return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'MKF'}";   
	&$debug("created $path\n");        
    }

    # set perms, because mkpath doesn't do it right (sometimes)
    chmod oct $perm, $path;
    &$debug("set permissions to $perm for $path\n");

    # set ownership
    chown $uid, $gid, $path;
    &$debug("set owner to $uid:$gid for $path\n");
    return "$CONFIG::API{'R_MKDIR'}:1";
}

# Dispatches mkdir call to appropriate host
sub default_remote_mk_dir ($$$$$) {
    return &RatsClientLib::send_remote_command("MKDIR",@_);
}

# get exclusive file lock. It takes a file handle as the only arg
sub GetFileLock ($) {
    if (flock($_[0], LOCK_EX)) {               # wait for lock
        seek($_[0], 0, 2);                     # make sure you have the
	return $CONFIG::RESPONSE{'SCS'};        # begining of the file
    }                                  
    return $CONFIG::RESPONSE{'FLF'};
}
 

# drop file lock for a given filehandle
sub GiveFileLock ($) {
    flock($_[0], LOCK_UN);     # and unlock
}

# sort unique of a given array
sub sort_unique (@) {
    my %saw_hash;
    @saw_hash{@_} =();
    return sort keys %saw_hash;
}

# return list of unique elements of a given list
sub unique (@) {
    my (%seen, @unique);
    @unique = grep { ! $seen{$_} ++ } @_;
    return @unique;
}

# this func takes as args a task ( "add"/"remove" ) and group name
# and a username. It returns SCS for success and some error for failure. 
sub group_edit ($$$) {
    my ($task, $grp, $usr, $grp, $pass, $task_gid, @members);
    my ($gname, $gpass, $gid, $members);
    my (@junk, $member, $entry, $old_entry, $count, $temp);
    my ($first_member, $response, $api_case, $empty_entry);
    
    ($task, $usr, $grp) = @_;

    # set some return api cases
    if ($task eq "add") {
	$api_case = 'R_ADDTOGRP';
    } elsif ($task eq "remove") {
	$api_case = 'R_REMFRGRP';
    }



    &$debug("Got $task $usr $grp\n");
    my $group_file = $CONFIG::ACCT_FILES{'group'};
    my $group_file_tmp = $CONFIG::ACCT_FILES_TMP{'group'};


    # note that if vgetgrnam succeds the usefull info is returned
    # other wise a error message gets stuffed into $grp so we 
    # need to loock for it.
    ($grp, $pass, $task_gid, $members) = vgetgrnam($grp);

    # if we didn't find the group then return and cry
    if ($grp eq $CONFIG::RESPONSE{'GRE'}) {
	return "$CONFIG::API{'CODEDERR'}:$grp";
    }
    if ($grp eq $CONFIG::RESPONSE{'NSG'}) {
	return "$CONFIG::API{'NOTFOUND'}:$grp";
    }

    # we lock the group file
    $response = &lock_file('group');
    unless ($response eq $CONFIG::RESPONSE{'SCS'}) {
	&$log("Failed to get group lock: $response!\n");
	return "$CONFIG::API{'CODEDERR'}:$response";
    }



    # We only care about entries which begin with the name of the
    # group and have the same gid. the rest we just write out
    # as soon as we see them

    open(GRP, $group_file) || 
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'GRE'}";
    open(TEMP, ">$group_file_tmp") || 
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'GTE'}";
    while (<GRP>) {
	chomp;
	($gname, $gpass, $gid, $temp) = split(":");
	unless ($gid == $task_gid && ($gname =~ /^$grp-/ || $gname eq $grp)) {
	    print TEMP $_,"\n";
	    next;
	}
    }
    close GRP;
   
    # Ok now .. we update the membership 
    @members = split(",", $members);
    if ($task eq "add") {
	push @members, $usr;
	@members = &unique(@members); # just in case it was already there
    } elsif ($task eq "remove") {
	@junk=();
	foreach $member (@members) {
	    push @junk, $member unless ($member eq $usr);
	}
	@members=@junk;
    }
    
    # and here we are supposed to update the group file
    # the problem is that we need to keep group length
    # under 1024 chars. We use <group>-ext<digit> groups
    # with the same gid as <group> for extra groups if
    # the entry is longer then 1024 chars. In fact because
    # we are paranoid we will keep entries under 900 chars
    # because we have no idea what YP may add to the entry
    # when it snarfs it.
    $entry = join(":", $grp, $pass, $task_gid);
    $empty_entry = $entry;
    $old_entry = $entry;
    $count=0;
    $first_member=1;
    foreach $member (@members) {
	if ($first_member) {
	    $entry = join(":", $old_entry, $member);
	    $first_member=0;
	} else {
	    $entry = join(",", $old_entry, $member);
	}
	# if we went over the we write the old one into the group file
	# and generate a new entry
	if ($CONFIG::GROUP_ENTRY_LENGTH_LIMIT < length($entry)) {
	    print TEMP $old_entry,"\n" ;
	    $count += 1;
	    $temp = $grp . "-" . $CONFIG::EXT_TOKEN . $count;
	    $old_entry = join(":", $temp, $pass, $task_gid, $member); 
	} else {
	    $old_entry = $entry;
	}
    }

    # check if no people where added to the entry and add
    # a : if so.
    if ($old_entry eq $empty_entry) {
	$old_entry = $old_entry . ":";
    }
    # now we only have the last chunk left to write down
    print TEMP $old_entry,"\n";
    
    close TEMP;
    
    #now we move the new file to the final location
    tmove ($group_file_tmp, $group_file);
    chmod oct $CONFIG::ACCT_FILES_PERMS{'group'}, $group_file;
    chown $CONFIG::ACCT_FILES_OWNER{'group'}, $CONFIG::ACCT_FILES_GROUP{'group'}, $group_file;
    &unlock_file('group');
    &$debug("Performed $task $usr $grp successfully\n");
    return "$CONFIG::API{$api_case}:1";
}

## this is the equivalent of 
## getgrnam, but it combines all extented groups into one answer
## for example getgrnam('ccfs') will only return members of ccfs
## but vgetgrnam('ccfs') will return members of ccfs and ccfs-<ext><digit>
sub vgetgrnam ($) {
    my $group_name = $_[0];
    my ($final_gn, $final_gp, $final_gid, $final_mem);
    my ($gn, $gp, $gid, $mem, $entry);
    my ($response);
    my $group_file = $CONFIG::ACCT_FILES{'group'};


    ($group_name) = split ("-$CONFIG::EXT_TOKEN", $group_name);
    &$debug("looking for $group_name\n");
    # read all of it
    open (GROUP, "$group_file") || return $CONFIG::RESPONSE{'GRE'};

    # find the gid of the group we are looking for
    foreach $entry (<GROUP>) {
	chomp $entry;
	($gn, $gp, $gid, $mem) = split(":", $entry);
	if ($gn eq $group_name) {
	    $final_gn = $gn;
	    $final_gid = $gid;
	    $final_gp = $gp;
	    $final_mem = $mem;
	    last;
	}
    }
    # if we didn't find it then we return an error
    unless ($final_gid) {
	return $CONFIG::RESPONSE{'NSG'};
    }
    # now find the members of the group and extentions
    seek(GROUP,0,0); #rewind the group file
    foreach $entry (<GROUP>) {
	chomp $entry;
	($gn, $gp, $gid, $mem) = split(":", $entry);
	if (($gn =~ /^$group_name-$CONFIG::EXT_TOKEN/) && $gid == $final_gid) {
	    $final_mem = join(",", $final_mem, $mem);
	}
    }
    close GROUP;

    
    # and we return it
    return ($final_gn, $final_gp, $final_gid, $final_mem);
}

# these are the main lock and unlock routines. They share a file handle 
{

my %lock_fh_hash;

# this routine tries to lock the group file using a fcntl
# lock on a lock file. Return SCS if it worked, error otherwise.
sub lock_file($) {
    my $file = $_[0];
    &$debug("Got $file\n");
    my $lock_file = $CONFIG::LOCK_FILE{$file};
    &$debug("Using $lock_file\n");
    my $try = 0;
    while ($try < $CONFIG::LOCK_TRIES) {
	if ( -e $lock_file) {
	    $try++;
	    sleep $CONFIG::SLEEP_LEN;
	    next;
	} else {
	    if ($lock_fh_hash{$file}) {
		&$debug("this proccess already locked $file\n");
		return $CONFIG::RESPONSE{'FAL'};
	    }
	    my $lock_fh = FileHandle->new(); 
	    $lock_fh_hash{$file} = $lock_fh;
	    open ($lock_fh, ">$lock_file") || return $CONFIG::RESPONSE{'ELF'};
	    my $response =  GetFileLock($lock_fh);
	    
	    # if we failed to get the fcntl then we nuke the file
	    unless ($response eq $CONFIG::RESPONSE{'SCS'}) {
		close $lock_fh;
		delete $lock_fh_hash{$file};
		unlink $lock_file;
	    }
	    
	    return $response;
	}
    }
    return $CONFIG::RESPONSE{'LTE'};
}

## this routine unlock the group file
sub unlock_file ($) {
    my $file = $_[0];
    my $lock_file = $CONFIG::LOCK_FILE{$file};
    if ($lock_fh_hash{$file}) {
	GiveFileLock($lock_fh_hash{$file});
	close $lock_fh_hash{$file};
	delete $lock_fh_hash{$file};
	unlink $lock_file;
    }
}
} # end of group lock routines.

# this routine moves the file given by arg1 to arg2 
sub tmove ($$) {
    my ($source, $target) =@_;
    # first we try the system way
    return $CONFIG::RESPONSE{'SCS'}  if rename $source, $target;
    # ok so that didn't work. So we do a copy and unlink
    # we should check return values here but what would I do 
    # with it upstream ?? TODO
    copy ($source, $target) ;
    # we are making the dangerous assumption that we do not
    # care if the unlink failed. We could check for that, but
    # I doubt we really give a RATS a**.
    unlink $source;
    return $CONFIG::RESPONSE{'SCS'};
}


# This routine figures out which host is responsible for group 
# updates and distpaches the add user to group request
# it takes username and group as args and returns the appropriate
# success or failure status
sub default_add_user_to_group ($$) {
  my ($usr,$grp) = @_;
  &$debug("Got $usr $grp\n");
  
  #check if the user has to many groups
  my $ret = &$do('GETGROUP', $usr, $CONFIG::ACCT_FILES_MASTER{'group'},$CONFIG::CLIENT_PORT);
  my ($code, @data) = split (":", $ret);
  if ($code == $CONFIG::API{'R_GETGROUP'}) {
    if (@data >= $CONFIG::MAX_GROUPS) {
      return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TMG'}";
    }
  } else {
    return $ret;
  } 
  
    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $grp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DADDTOGRP'};
            return &$handler_sub(@_);
        # if remote then set $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    # figure where the group file resides
    } elsif ($CONFIG::ACCT_FILES_MASTER{'group'} eq $host) {
        &$debug("spawn local $usr $grp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LADDTOGRP'};
        return &$handler_sub(@_);
    } else {
        $target_host = $CONFIG::ACCT_FILES_MASTER{'group'};
    }


    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $grp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RADDTOGRP'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}

# adds an user to a group in the local DB
sub default_db_add_user_to_group ($$) {
    my ($usr, $grp) = @_;
    use DBI;
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth, $gid, $count,$lock,$timeout);

    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);

    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }

    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"group_file-$grp\",\"$timeout\")") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});

    $sth->execute() ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
	$sth->finish();
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
	$sth->finish();
	$dbh->disconnect();
    	return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }


    # get gid for the group we are looking for
    $sth = $dbh->prepare("SELECT gid FROM group_file WHERE gname = \"$grp\"") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || 
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns(\$gid);
    $sth->fetch();
    unless ($gid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSG'}";
    }
    $sth->finish();

    # check if the username/group pair already exits
    $sth = $dbh->prepare("SELECT COUNT(*) FROM group_members WHERE uname = \"$usr\" AND gid = $gid")||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});

    $sth->execute()  ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});

    $count = $sth->fetchrow_array();
    $sth->finish();

    # return if user already in group
    if ($count) {
	$dbh->disconnect();
	return "$CONFIG::API{'R_ADDTOGRP'}:1";
    }

    
    # insert the new entry
    $dbh->do("INSERT group_members VALUES($gid, \"$usr\")") ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    

    $dbh->disconnect();
    return "$CONFIG::API{'R_ADDTOGRP'}:1";
}

# Dispatches add to group call to appropriate host
sub default_remote_add_user_to_group (@) {
    return &RatsClientLib::send_remote_command("ADDTOGRP",@_);
}

# calls the appropriate functions for local group entry addition
sub default_local_add_user_to_group (@) {
    return &group_edit("add", @_);
}

# This routine figures out which host is responsible for group 
# updates and distpaches the remove user from group request
# it takes username and group as args and returns the appropriate
# success or failure status
sub default_remove_user_from_group ($$) {
    my ($usr,$grp) = @_;

    &$debug("Got $usr $grp\n");

    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $grp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DREMFRGRP'};
            return &$handler_sub(@_);
        # if remote then set $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    # figure where the group file resides
    } elsif ($CONFIG::ACCT_FILES_MASTER{'group'} eq $host) {
        &$debug("spawn local $usr $grp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LREMFRGRP'};
        return &$handler_sub(@_);
    } else {
        $target_host = $CONFIG::ACCT_FILES_MASTER{'group'};
    }


    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $grp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RREMFRGRP'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}



# remove an user from a group in the local DB
sub default_db_remove_user_from_group ($$) {
    my ($usr, $grp) = @_;
    use DBI;
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth, $gid, $count, $timeout, $lock);

    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);

    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});


    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }

   # lock entry while we muck around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"group_file-$grp\",\"$timeout\")") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});

    $sth->execute() ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
	$sth->finish();
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
	$sth->finish();
	$dbh->disconnect();
    	return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }

    # get gid for the group we are looking for
    $sth = $dbh->prepare("SELECT gid FROM group_file WHERE gname = \"$grp\"") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || 
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns(\$gid);
    $sth->fetch();
    unless ($gid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSG'}";
    }
    $sth->finish();

    # check if the username/group pair already exits
    $sth = $dbh->prepare("SELECT COUNT(*) FROM group_members WHERE uname = \"$usr\" AND gid = $gid")||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});

    $sth->execute()  ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});

    $count = $sth->fetchrow_array();
    $sth->finish();

    # return if user already deleted from  group
    unless ($count) {
	$dbh->disconnect();
	return "$CONFIG::API{'R_REMFRGRP'}:1";
    }

    
    # insert the new entry
    $dbh->do("DELETE FROM group_members WHERE gid = $gid AND uname = \"$usr\"") ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    

    $dbh->disconnect();
    return "$CONFIG::API{'R_REMFRGRP'}:1";
}



# Dispatches add to group call to appropriate host
sub default_remote_remove_user_from_group ($$$) {
    return &RatsClientLib::send_remote_command("REMFRGRP",@_);
}

# calls the appropriate functions for local group entry removal
sub default_local_remove_user_from_group ($$) {
    return &group_edit("remove", @_);
}

# remove a given path
sub default_rm_dir ($) {
    my $path = $_[0];
    # figure our where the actual fs is 
    # or if it even exists
    &$debug("Got $path\n");
    my $fs = &RatsClientLib::determine_fs($path);
    if ($fs eq "") {
	&$log("Can't determine fs for $path\n");
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NFS'}";
    }
    &$debug("fs for $path is $fs\n");
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");
    my $target_host = $CONFIG::FS{$fs};
    if ($host eq $target_host) {
	&$debug("spawn local $path\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'LRMDIR'};
	return &$handler_sub(@_);
    } else {
	&$debug("spawn remote $path\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'RRMDIR'};
	return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);
    }
}

# this function takes a path and remove it
sub default_local_rm_dir ($) {
    my $path = $_[0];
    &$debug("Got $path\n");

    # check path and untaint
    if ($path =~ /^($CONFIG::PATH_REGEXP)$/) {
        $path = $1;                     # $path now untainted
    } else {
        &$log("Bad data in $path\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }

    # if it exists remove it
    if (-d $path) {
	&vrmtree($path, 0, 0) || return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'RMF'}";
    }
    return "$CONFIG::API{'R_RMDIR'}:1";
}


# Dispatches rmdir call to appropriate host
sub default_remote_rm_dir ($$$$$) {
    return &RatsClientLib::send_remote_command("RMDIR",@_);
}


# This routine figures out which host is responsible for password
# updates and distpaches the add entry request
# it takes the pw/shadow  entry parts as args and returns the appropriate
# success or failure status
sub default_add_pw_entry (@) {

    my ($usr,$pass,$uid,$gid,$gcos,$home,$shell,$lastchg,$mindays,$maxdays,$warn,$inactive,$expire,$flag) = @_;
    &$debug("Got $usr $pass $uid $gid $gcos $home $shell $lastchg $mindays $maxdays $warn $inactive $expire $flag\n");

    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
	# if local then do it
	if ($CONFIG::CLIENT_DB_HOST eq $host) {
	    &$debug("spawn DB $usr $pass $gid $gcos $home $shell\n");
	    my $handler_sub = $CONFIG::COMMAND_HOOKS{'DADDPWEN'};
	    return &$handler_sub(@_);
	# if remote then set $target_host
	} else {
	    $target_host = $CONFIG::CLIENT_DB_HOST;
	}
    # are we doing a local file password add?
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
	&$debug("spawn local $usr $pass $gid $gcos $home $shell\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'LADDPWEN'};
	return &$handler_sub(@_);
    # are we doing a remote password add to a file?
    } else {
	# figure our where the actual password file is maintained
	$target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }
	

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $pass $gid $gcos $home $shell\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RADDPWEN'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);
}


# this routine adds a password and shadow entry to the local db.
# it takes the pw/shadow  entry parts as args and returns the appropriate
# success or failure status
sub default_db_add_pw_entry (@) {
    my ($usr,$pass,$uid,$gid,$gcos,$home,$shell,$lastchg,$mindays,$maxdays,$warn,$inactive,$expire,$flag) = @_;

    &$debug("Got $usr $pass $uid $gid $gcos $home $shell $lastchg $mindays $maxdays $warn $inactive $expire $flag\n");

    use DBI;
    use strict;

    # setup DB vars
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth, @pass_parts);
    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);
    

    my ($old_uname, $old_pass, $old_uid, $old_gid,$old_gcos, $old_home);
    my ($old_shell, $old_last, $old_min, $old_max, $old_warn,$old_inactive);
    my ($old_exp, $old_flag, $old_chg, $old_active, $lock, $timeout);

    # reserve uid if required
    my $host = &get_hostname();
    # get remote, first try to look it up, then reserve it
    if ($CONFIG::UID_SYNC && $CONFIG::UID_HOST ne $host) {
	&$debug("trying to obtained remote uid for $usr\n");
	#getuid from remote host
	my $result = &$do('GETUID', $usr, $CONFIG::UID_HOST, $CONFIG::CLIENT_PORT);
	my ($code, @data) = split(":", $result);
	if ($code == $CONFIG::API{'R_GETUID'}) {      # try to get it for existing user
	    $uid = $data[0];
	} elsif ($code == $CONFIG::API{'NOTFOUND'}) { # reserve uid
	    &$debug("trying to reserve remote uid for $usr\n");
	    $result = &$do('RESERVE', $usr, $CONFIG::UID_HOST, $CONFIG::CLIENT_PORT);
	    ($code, @data) = split(":", $result);
	    if ($code == $CONFIG::API{'R_RESERVE'}) { 
		$uid = $data[0];
	    } else {
		&$log("failed to get uid for $usr: @data\n");
		return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'UGF'}";
	    }
	} else {
	    &$log("failed to get uid for $usr: @data\n");
	    return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'UGF'}";
	}
    }

    # DB uid's and gid's are offset by 1
    $uid++; $gid++;


    # connect
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) || 
	return &mysql_error($CONFIG::RESPONSE{'DCF'});


    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }



    # lock entry while we muck around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$usr\",\"$timeout\")") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});

    $sth->execute() ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
	$sth->finish();
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
	$sth->finish();
	$dbh->disconnect();
    	return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }

    # check if data already exists in DB
    $sth  = $dbh->prepare("SELECT * FROM password_file WHERE uname = ? OR uid = ?") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute($usr, $uid) ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$old_uname, \$old_pass, \$old_uid, \$old_gid,\$old_gcos, \$old_home,
			\$old_shell, \$old_last, \$old_min, \$old_max, \$old_warn,
			\$old_inactive, \$old_exp, \$old_flag, \$old_chg, \$old_active);


    while ($sth->fetch()) {
	# first check if user already active
	if ($old_uname eq $usr && $old_active eq "1") {
	    $sth->finish();
	    $dbh->disconnect();
	    return "$CONFIG::API{'ITEXISTS'}:$CONFIG::RESPONSE{'UAE'}";
	}
	# check if uid already taken
	if ($old_uid == $uid && $old_uname ne $usr) {
	    $sth->finish();
	    $dbh->disconnect();
	    return "$CONFIG::API{'ITEXISTS'}:$CONFIG::RESPONSE{'UAE'}:";
	}
	# check if new uid and old inactive uid don't match, unless 
	# we need to generate a new uid for user
	if ($old_uname eq $usr && $old_uid != $uid && $uid != $CONFIG::BOGUS_UID + 1) {
	    $sth->finish();
	    $dbh->disconnect();
	    return "$CONFIG::API{'CODEDERR'}:Existing inactive and new UID dont match";
	}
	# if an uid alread exists and user did not provide one then use the old.
	if ($old_uname eq $usr && $uid == $CONFIG::BOGUS_UID + 1) {
	    $uid = $old_uid;
	}
    }    
    $sth->finish();
    
    # insert the new entry. 
    #We use replace to ensure override of inactive entries
    $sth = $dbh->prepare( qq{
	REPLACE password_file VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)
	    }) || 
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    
    $sth->execute("$usr","$pass",$uid,$gid,"$gcos","$home","$shell",$lastchg,$mindays,$maxdays,$warn,$inactive,$expire,$flag,undef,"1") ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    
    $sth->finish();
    
    
    # if we had a null uid fed to the db, then we should check 
    # what it was set to.
    unless ($uid) {
	$uid = $dbh->{'mysql_insertid'};
    }
    # and normalize it.
    $uid--;

    #clean up
    $dbh->disconnect();


    #return success
    return "$CONFIG::API{'R_ADDPWEN'}:$uid";
}  



# Dispatches add to pw call to appropriate host
sub default_remote_add_pw_entry (@) {
    return &RatsClientLib::send_remote_command("ADDPWEN",@_);
}

# this function takes password entry parts, merges them, and adds
# the to a password file, shadow file and so on
# it returns SCS on success or the appropriate error message otherwise
sub default_local_add_pw_entry (@) {

    # arg!! the number of args!
    my ($usr,$pass,$uid,$gid,$gcos,$home,$shell,$lastchg,$mindays,$maxdays,$warn,$inactive,$expire,$flag) = @_;    
    my (@passwd, $entry, $pass_file, $shadow_file, @shadow);
    my ($temp_user, @rest, $handler_sub, $temp_passwd, $temp_shadow);
    my ($response, $temp_pass, $temp_uid, @reserve);
    my ($res_file, $temp_res, $already_reserved, $temp_entry);
    

    # now we lock various files
    $response = &lock_file('passwd');
    unless ($response eq $CONFIG::RESPONSE{'SCS'}) {
	&$log("Failed passwd lock: $response!\n");
	return "$CONFIG::API{'CODEDERR'}:$response";
    }
    $response = &lock_file('shadow');
    unless ($response eq $CONFIG::RESPONSE{'SCS'}) {
	&unlock_file('passwd');
	&$log("Failed shadow lock: $response!\n");
	return "$CONFIG::API{'CODEDERR'}:$response";
    }

    #lock the reserve file
    $res_file = $CONFIG::ACCT_FILES{'reserve'};
    $response = &lock_file('reserve');
    unless ($response eq $CONFIG::RESPONSE{'SCS'}) {
	&unlock_file('passwd');
	&unlock_file('shadow');
	&$log("Failed reserve lock: $response!\n");
	return "$CONFIG::API{'CODEDERR'}:$response";
    }
    

    # now we read the whole password file into memory
    # this is sort of wastfull. Maybe we should just run through it
    $pass_file = $CONFIG::ACCT_FILES{'passwd'};
    unless (open(PASSWD, "$pass_file")) {
	&unlock_file('shadow');
	&unlock_file('passwd');
	&unlock_file('reserve');
	&$log("Failed passwd open: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'PRE'}";
    }
    @passwd = <PASSWD>;
    close PASSWD;

    # now we read the whole shadow file into memory
    $shadow_file = $CONFIG::ACCT_FILES{'shadow'};
    unless (open(SHADOW, "$shadow_file")) {
	&unlock_file('shadow');
	&unlock_file('passwd');
	&unlock_file('reserve');
	&$log("Failed shadow open: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'SRE'}";
    }
    @shadow = <SHADOW>;
    close SHADOW;

    # now we read the whole reserve file into memory
    $res_file = $CONFIG::ACCT_FILES{'reserve'};
    if (-f $res_file) {
	unless (open(RES, "$res_file")) {
	    &unlock_file('shadow');
	    &unlock_file('passwd');
	    &unlock_file('reserve');
	    &$log("Failed reserve open: $!\n");
	    return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'RRE'}";
	}
	@reserve = <RES>;
	close RES;
    }

    # first we verify that the user does not already exist
    # or the uid was already used
    foreach $entry (@passwd) {
	($temp_user, $temp_pass, $temp_uid, @rest) = split(":", $entry);
	if (($temp_user eq $usr) || ($temp_uid == $uid)) {
	    &unlock_file('shadow');
	    &unlock_file('passwd');
	    &unlock_file('reserve');
	    &$log("$usr or uid $uid already exists\n");
	    return "$CONFIG::API{'ITEXISTS'}:$CONFIG::RESPONSE{'UAE'}";
	}
    }

    # check if the uid was already reserved for this user
    # or a different one
    foreach $entry (@reserve) {
	$temp_entry = $entry;
	chomp $temp_entry;
	($temp_user, $temp_uid) = split(":", $temp_entry);
	if (($temp_user eq $usr) && ($temp_uid != $uid) && ($uid != $CONFIG::BOGUS_UID)) {
	    &unlock_file('shadow');
	    &unlock_file('passwd');
	    &unlock_file('reserve');
	    &$log("$usr or uid $uid already exists\n");
	    return "$CONFIG::API{'ITEXISTS'}:$CONFIG::RESPONSE{'UAE'}";
	} elsif (($temp_user eq $usr) && ($uid == $CONFIG::BOGUS_UID)) {
	    $already_reserved = 1;
	    $uid = $temp_uid;
	} elsif (($temp_user ne $usr) && ($uid == $temp_uid)) {
	    &unlock_file('shadow');
	    &unlock_file('passwd');
	    &unlock_file('reserve');
	    &$log("$usr or uid $uid already exists\n");
	    return "$CONFIG::API{'ITEXISTS'}:$CONFIG::RESPONSE{'UAE'}";
	}
    }

    # now that we have the the file locked we can generate a UID.
    # if the uid is the bogus (according to config file) then we 
    # generate one. Otherwise we use the  given one
    if ($uid == $CONFIG::BOGUS_UID) {
	$handler_sub = $CONFIG::COMMAND_HOOKS{'GENUID'};
	$uid = &$handler_sub($usr);
	# check if it failed
	if ($uid == $CONFIG::BOGUS_UID) {
	    &unlock_file('shadow');
	    &unlock_file('passwd');
	    &unlock_file('reserve');
	    return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'UGF'}";
	}
	# we mark this because GENUID reserves the uid
	$already_reserved = 1;
    }

    # now we can generate the password entry
    $entry = join(":", $usr,"x",$uid,$gid,$gcos,$home,$shell);
    
    # write out the new password file to a temp file
    $temp_passwd = $CONFIG::ACCT_FILES_TMP{'passwd'};
    unless (open(PW_TEMP, ">$temp_passwd")) {
	&unlock_file('shadow');
	&unlock_file('passwd');
	&unlock_file('reserve');
	&$log("Failed passwd temp open: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'PTE'}";
    }
    print PW_TEMP @passwd;
    print PW_TEMP "$entry\n";
    close PW_TEMP;
    
    #update reserve file if not already
    unless ($already_reserved) {
	# write out the new reserve file to a temp file
	$temp_res = $CONFIG::ACCT_FILES_TMP{'reserve'};
	unless (open(RS_TEMP, ">$temp_res")) {
	    &unlock_file('shadow');
	    &unlock_file('passwd');
	    &unlock_file('reserve');
	    &$log("Failed reserve temp open: $!\n");
	    return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'RTE'}";
	}
	print RS_TEMP @reserve;
	print RS_TEMP "$usr:uid\n";
	close RS_TEMP;
	tmove ($temp_res, $res_file);
	chmod oct $CONFIG::ACCT_FILES_PERMS{'reserve'}, $res_file;
	chown $CONFIG::ACCT_FILES_OWNER{'reserve'}, $CONFIG::ACCT_FILES_GROUP{'reserve'}, $res_file;
    }

    # now we generate the shadow entry 
    $entry = join(":", $usr,$pass,$lastchg,$mindays,$maxdays,$warn,$inactive,$expire,$flag);

    # write out the new shadow file to a temp file
    $temp_shadow = $CONFIG::ACCT_FILES_TMP{'shadow'};
    unless (open(SH_TEMP, ">$temp_shadow")) {
	&unlock_file('shadow');
	&unlock_file('passwd');
	&unlock_file('reserve');
	&$log("Failed passwd temp open: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'STE'}";
    }
    print SH_TEMP @shadow;
    print SH_TEMP "$entry\n";
    close SH_TEMP;

    # and we move them in the right place
    tmove ($temp_passwd, $pass_file);
    chmod oct $CONFIG::ACCT_FILES_PERMS{'passwd'}, $pass_file;
    chown $CONFIG::ACCT_FILES_OWNER{'passwd'}, $CONFIG::ACCT_FILES_GROUP{'passwd'}, $pass_file;
    tmove ($temp_shadow, $shadow_file);
    chmod oct $CONFIG::ACCT_FILES_PERMS{'shadow'}, $shadow_file;
    chown $CONFIG::ACCT_FILES_OWNER{'shadow'}, $CONFIG::ACCT_FILES_GROUP{'shadow'}, $shadow_file;

    # and unlock both
    &unlock_file('shadow');
    &unlock_file('passwd');
    &unlock_file('reserve');
    &$log("$usr added to passwd/shadow\n");
    return "$CONFIG::API{'R_ADDPWEN'}:$uid";
}

# the default generate uid function
# WARNING: this function assumes that you already have the password file
# WARNING: locked. Otherwise multiple proccess could get the same uid 
# WARNING: before the first one finished updating the file
#
# it first checks if UID syncronization if being done. If so it contacts 
# the appropriate remote host. If this is running on a UID sync master 
# then it generates a new uid.

sub default_generate_uid  ($) {
    my ($result, $code, @data, $handler_sub, $remote, $port);
    my ($res_file, $uid, $usr);
    my $user = $_[0];
    my $host = &get_hostname();

    # get remote, first try to look it up, then reserve it
    if ($CONFIG::UID_SYNC && $CONFIG::UID_HOST ne $host) {
	&$debug("trying to obtained remote uid for $user\n");
	#getuid from remote host
	$result = &$do('GETUID', $user, $CONFIG::UID_HOST, $CONFIG::CLIENT_PORT);
	($code, @data) = split(":", $result);
	if ($code == $CONFIG::API{'R_GETUID'}) {      # try to get it for existing user
	    return $data[0];
	} elsif ($code == $CONFIG::API{'NOTFOUND'}) { # reserve uid
	    &$debug("trying to reserve remote uid for $user\n");
	    $result = &$do('RESERVE', $user, $CONFIG::UID_HOST, $CONFIG::CLIENT_PORT);
	    ($code, @data) = split(":", $result);
	    if ($code == $CONFIG::API{'R_RESERVE'}) { 
		return $data[0];
	    } else {
		&$log("failed to reserve uid for $user: @data\n");
		return $CONFIG::BOGUS_UID;
	    }
	} else {
	    &$log("failed to get uid for $user: @data\n");
	    return $CONFIG::BOGUS_UID;
	}

    } else {    # get local

	&$debug("trying to obtained local uid for $user\n");
	# first we check if was already reserved
	$res_file = $CONFIG::ACCT_FILES{'reserve'};
	if ( -f $res_file) {
	    unless (open(RESERVE, $res_file)) {
		&$log("Failed to open reserve file: $!\n");
		return $CONFIG::BOGUS_UID;
	    }
	    while (<RESERVE>) {
		chomp;
		($usr, $uid) = split(":", $_);
		if ($user eq $usr) {
		    return $uid;
		}
	    }
	    close RESERVE;
	}
	# otherwise we generate one
	$handler_sub = $CONFIG::COMMAND_HOOKS{'GENNUID'};
	return &$handler_sub($user);
    }
	
}


# This routine figures out which host is responsible for password
# updates and distpaches the remove entry request
# it takes the username as arg and returns the appropriate
# success or failure status
sub default_remove_pw_entry ($) {
    my $usr = $_[0];
    &$debug("Got $usr\n");

    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
	# if local then do it
	if ($CONFIG::CLIENT_DB_HOST eq $host) {
	    &$debug("spawn DB $usr\n");
	    my $handler_sub = $CONFIG::COMMAND_HOOKS{'DREMPWEN'};
	    return &$handler_sub(@_);
	# if remote then sety $target_host
	} else {
	    $target_host = $CONFIG::CLIENT_DB_HOST;
	}
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
	&$debug("spawn local $usr\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'LREMPWEN'};
	return &$handler_sub(@_);
    #
    } else {
	# 
	$target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }
	

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RREMPWEN'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);
}



# remove an user from the DB password file. BTW, all this
# does is update the entry to set the user to inactive.
sub default_db_remove_pw_entry (@_) {
    my $user = $_[0];
    use DBI;

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth, $count,$timeout,$lock);
    
    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);
    
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});
    
    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }

    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$user\",\"$timeout\")") ||
	return &mysql_error("$CONFIG::RESPONSE{'DPF'}");

    $sth->execute() ||
	return &mysql_error("$CONFIG::RESPONSE{'DEF'}");
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
	$sth->finish();
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
	$sth->finish();
	$dbh->disconnect();
    	return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }

    # first we check if the user exists
    $sth = $dbh->prepare("SELECT COUNT(*) FROM password_file WHERE uname = \"$user\" AND active = \"1\"")||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});

    $sth->execute()  ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});

    $count = $sth->fetchrow_array();

    $sth->finish();
    unless ($count) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }


    $sth = $dbh->do("UPDATE password_file SET active = \"0\" WHERE uname = \"$user\"") ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});

    $dbh->disconnect();
    return "$CONFIG::API{'R_REMPWEN'}:1";
}



# Dispatches remove from pw call to appropriate host
sub default_remote_remove_pw_entry (@) {
    return &RatsClientLib::send_remote_command("REMPWEN",@_);
}


# this function takes a username and remove the passwd/shadow entries for it
# it returns SCS on success or the appropriate error message otherwise
sub default_local_remove_pw_entry (@) {

    my $usr = $_[0];
    my (@passwd, $entry, $pass_file, $shadow_file, @shadow);
    my ($temp_user, @rest, $temp_passwd, $temp_shadow);
    my ($response, $found);

    # now we lock various files
    $response = &lock_file('passwd');
    unless ($response eq $CONFIG::RESPONSE{'SCS'}) {
	&$log("Failed passwd lock: $response!\n");
	return "$CONFIG::API{'CODEDERR'}:$response";
    }
    $response = &lock_file('shadow');
    unless ($response eq $CONFIG::RESPONSE{'SCS'}) {
	&unlock_file('passwd');
	&$log("Failed shadow lock: $response!\n");
	return "$CONFIG::API{'CODEDERR'}:$response";
    }

    # now we read the whole password file into memory
    $pass_file = $CONFIG::ACCT_FILES{'passwd'};
    unless (open(PASSWD, "$pass_file")) {
	&unlock_file('shadow');
	&unlock_file('passwd');
	&$log("Failed passwd open: $!\n");
	return"$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'PRE'}";
    }
    @passwd = <PASSWD>;
    close PASSWD;

    # now we read the whole shadow file into memory
    $shadow_file = $CONFIG::ACCT_FILES{'shadow'};
    unless (open(SHADOW, "$shadow_file")) {
	&unlock_file('shadow');
	&unlock_file('passwd');
	&$log("Failed shadow open: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'SRE'}";
    }
    @shadow = <SHADOW>;
    close SHADOW;
    
    # first we verify that the user exist
    $found = 0;
    foreach $entry (@passwd) {
	($temp_user, @rest) = split(":", $entry);
	if ($temp_user eq $usr) {
	    $found = 1;
	}
    }
    unless ($found) {
	&unlock_file('shadow');
	&unlock_file('passwd');
	&$log("$usr does not exist\n");
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }


    # write out the new password file to a temp file
    $temp_passwd = $CONFIG::ACCT_FILES_TMP{'passwd'};
    unless (open(PW_TEMP, ">$temp_passwd")) {
	&unlock_file('shadow');
	&unlock_file('passwd');
	&$log("Failed passwd temp open: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'PTE'}";
    }
    foreach $entry (@passwd) {
	($temp_user, @rest) = split(":", $entry);
	print PW_TEMP $entry unless ($temp_user eq $usr);
    }
    close PW_TEMP;
    
    # write out the new shadow file to a temp file
    $temp_shadow = $CONFIG::ACCT_FILES_TMP{'shadow'};
    unless (open(SH_TEMP, ">$temp_shadow")) {
	&unlock_file('shadow');
	&unlock_file('passwd');
	&$log("Failed passwd temp open: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'STE'}";
    }
    foreach $entry (@shadow) {
	($temp_user, @rest) = split(":", $entry);
	print SH_TEMP $entry unless ($temp_user eq $usr);
    }
    close SH_TEMP;

    # and we move them in the right place
    tmove ($temp_passwd, $pass_file);
    chmod oct $CONFIG::ACCT_FILES_PERMS{'passwd'}, $pass_file;
    chown $CONFIG::ACCT_FILES_OWNER{'passwd'}, $CONFIG::ACCT_FILES_GROUP{'passwd'}, $pass_file;
    tmove ($temp_shadow, $shadow_file);
    chmod oct $CONFIG::ACCT_FILES_PERMS{'shadow'}, $shadow_file;
    chown $CONFIG::ACCT_FILES_OWNER{'shadow'}, $CONFIG::ACCT_FILES_GROUP{'shadow'}, $shadow_file;

    # and unlock both
    &unlock_file('shadow');
    &unlock_file('passwd');
    &$debug("$usr removed from passwd/shadow\n");
    return "$CONFIG::API{'R_REMPWEN'}:1";
}


# this routine get a username and returns the uid for that user.
# if the username does not exist it returns an error. It does
# this by distpaching the call to the local or remote machine
sub default_get_uid ($) {
    my $usr = $_[0];

    &$debug("Got $usr\n");
    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DGETUID'};
            return &$handler_sub(@_);
        # if remote then set $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
        &$debug("spawn local $usr\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LGETUID'};
        return &$handler_sub(@_);
    #
    } else {
        # 
        $target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }
        

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RGETUID'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}

# get uid for an user from the local DB
sub default_db_get_uid (@) {
    my $usr = $_[0];
        use DBI;

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);
    my ($uid, $lastchange);

    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);
    
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});
    
    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # check if user already exists
    $sth = $dbh->prepare("SELECT uid,lastchange FROM password_file WHERE uname = \"$usr\"") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid,\$lastchange);
    $sth->fetch();
    $sth->finish();
    $dbh->disconnect();

    if ($uid) {
	$uid--;
	return "$CONFIG::API{'R_GETUID'}:$uid:$lastchange";
    } else {
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
}


# Dispatches get uid call to appropriate host
sub default_remote_get_uid (@) {
    return &RatsClientLib::send_remote_command("GETUID",@_);
}

# determine the uid for a given username from local passwd file
# return uid if succesfull, error if failed
sub default_local_get_uid ($) {
    my ($uid, $entry, $pass_file, @passwd, $res_file);
    my ($usr, $pass, @rest, $target, $val, $found);
    $target = $_[0];
    $found = 0;
    $val = $CONFIG::BOGUS_UID;
    # read the passwd file, we don't care is locked
    # maybe we should?
    &$debug("got $target \n");
    $pass_file = $CONFIG::ACCT_FILES{'passwd'};
    unless (open(PASSWD, $pass_file)) {
	&$log("Failed to open passwd file: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'PRE'}";
    }
    while (<PASSWD>) {
	chomp;
	($usr, $pass, $uid, @rest) = split(":", $_);
	if ($target eq $usr) {
	    $val = $uid;
	    $found = 1;
	} 
    }
    close PASSWD;

    # check the reserved file
    $res_file = $CONFIG::ACCT_FILES{'reserve'};
    if ( -f $res_file && !$found) {
	unless (open(RESERVE, $res_file)) {
	    &$log("Failed to open reserve file: $!\n");
	    return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'RRE'}";
	}
	@passwd = <RESERVE>;
	close RESERVE;
	foreach $entry (@passwd) {
	    chomp $entry;
	    ($usr, $uid) = split(":", $entry);
	    if ($target eq $usr) {
		$val = $uid;
	    }
	}
    }
    

    &$debug("returning $val\n");
    if ($val == $CONFIG::BOGUS_UID) {
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    } else {
	# we add the 0 for the non-existant time stamp supported by DB
	return "$CONFIG::API{'R_GETUID'}:$val:0";
    }


}


# runs a script 
sub default_run_script ($$) {
    my ($script, $user) = @_;
    &$debug("got $script\n");

    # check $script and untaint
    if ($script =~ /^($CONFIG::PATH_REGEXP)$/) {
        $script = $1;                     # $script now untainted
    } else {
        &$log("Bad data in $script\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }  

    # check $user and untaint
    if ($user =~ /^([\w]+)$/) {
        $user = $1;                     # $user now untainted
    } else {
        &$log("Bad data in $user\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }  

    # now we prepand the script directory path
    $script = join("/", $CONFIG::SCRIPT_DIR, $script);
    if ( -f $script ) {
	&$debug("trying to run $script\n");
        system ($script,$user) or return "$CONFIG::API{'CODEDERR'}:$!}";
	&$debug("ran $script\n");
	return "$CONFIG::API{'R_RUN'}:1";
    } else {
	&$log("no such file $script\n");
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NFL'}";
    }
}



# get pwent. takes username returns passwd entry or error
sub default_getpwent($) {
    my $usr = $_[0];
    &$debug("Got $usr\n");

    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");
    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
	# if local then do it
	if ($CONFIG::CLIENT_DB_HOST eq $host) {
	    &$debug("spawn DB $usr\n");
	    my $handler_sub = $CONFIG::COMMAND_HOOKS{'DGETPWENT'};
	    return &$handler_sub(@_);
	# if remote then sety $target_host
	} else {
	    $target_host = $CONFIG::CLIENT_DB_HOST;
	}
    # are we doing a local file password lookup?
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
	&$debug("spawn local $usr\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'LGETPWENT'};
	return &$handler_sub(@_);
    # are we doing a remote lookup?
    } else {
	# figure our where the actual password file is maintained
	$target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }
	

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RGETPWENT'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);
}


# the DB version of getpwent
sub default_db_getpwent(@) {
    my $user = $_[0];
    use DBI;
    
    &$debug("got $user\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth, @pass_parts);

    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);

    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }

# there should be a max of one row, because usernames are unique
    $sth = $dbh->prepare("SELECT * FROM password_file WHERE uname = \"$user\" AND active = \"1\"") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});

    # there should be a max of one row, because usernames are unique
    my @data = $sth->fetchrow_array();
    
    $sth->finish();
    $dbh->disconnect();


    if (@data) {
	$data[2]--;$data[3]--;
	my $temp = join(":",@data[0 .. 6],$data[14]);
	return "$CONFIG::API{'R_GETPWENT'}:$temp";
    } else {
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
}


# Dispatches getpwent call to appropriate host
sub default_remote_getpwent (@) {
    return &RatsClientLib::send_remote_command("GETPWENT",@_);
}

# get pwent. takes username returns passwd entry or error
sub default_local_getpwent ($) {
    my ($uid, $entry, $pass_file, @passwd);
    my ($usr, $pass, @rest, $target, $val);
    $target = $_[0];

    $val = "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    # read the passwd file, we don't care is locked
    # maybe we should?
    &$debug("got $target \n");
    $pass_file = $CONFIG::ACCT_FILES{'passwd'};
    unless (open(PASSWD, $pass_file)) {
	&$log("Failed to open passwd file: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'PRE'}";
    }
    while (<PASSWD>) {
	chomp;
	($usr, $pass, $uid, @rest) = split(":", $_);
	if ($target eq $usr) {
	    # we add the 0 at then end as a timestamp placeholder
	    $val = "$CONFIG::API{'R_GETPWENT'}:$_:0";
	} 
    }
    close PASSWD;

    &$debug("returning $val\n");
    return $val;
}

# take uid and fs and return quotas for that user on that fs
# dispatch call as appropriate. returns soft and hard quotas
sub default_get_quota ($$) {
    my ($uid,$fs) = @_;

    &$debug("Got $uid $fs\n");

    # check $fs and untaint
    if ($fs =~ /^($CONFIG::PATH_REGEXP)$/) {
        $fs = $1;                     # $perm now untainted
    } else {
        &$log("Bad data in $fs\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }  

    my $new_fs = &RatsClientLib::determine_fs($fs);
    if ($new_fs eq "") {
	&$debug("Can't determine fs for $fs\n");
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NFS'}";
    } else {
	$fs = $new_fs;
    }
    
    my $host = &get_hostname();
    my $target_host = $CONFIG::FS{$fs};
    if ($host eq $target_host) {
	&$debug("spawn local $uid  $fs\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'LQGET'};
	return &$handler_sub($uid,$fs);
    } else {
	&$debug("spawn remote $uid $fs $target_host\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'RQGET'};
	return &$handler_sub($uid,$fs,$target_host,$CONFIG::CLIENT_PORT);
    }
}


# distpatch get quota call to remote host
sub default_remote_get_quota ($$$) {
    return &RatsClientLib::send_remote_command("QGET",@_);
}


# locally determine the quota for a given uid on a given fs
# and return soft and hard
sub default_local_get_quota ($$) {
    my ($uid,$fs) = @_;
    my $return;
    &$debug("Got $uid $fs\n");
    
    my $quotafile = join("/", $fs, "quotas");
    # check user and untaint
    if ($uid =~ /^(\d+)$/) {
        $uid = $1;                     # $uid now untainted
    } else {
        &$log("Bad data in $uid\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }
    
    # devine device
    my $dev =  Quota::getqcarg($fs);
    
    # get the quota
    my ($block_curr, $block_soft, $block_hard, $block_timelimit, $inode_curr, $inode_soft, $inode_hard, $inode_timelimit) = Quota::query($dev, $uid);
    
    #check error status
    my $error = Quota::strerr();
    unless ($error eq "error \#0") {
	return "$CONFIG::API{'CODEDERR'}:$error";
    }
    return "$CONFIG::API{'R_QGET'}:$block_soft:$block_hard";

    
}

# copy one file to another
sub default_copy ($$$$) {
    my ($source, $target, $uid, $gid) = @_;

    &$debug("got $source $target $uid $gid\n");

    # check $source and untaint
    if ($source =~ /^($CONFIG::PATH_REGEXP)$/) {
        $source = $1;                     # $source now untainted
    } else {
        &$log("Bad data in $source\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }  

    # check $target and untaint
    if ($target =~ /^($CONFIG::PATH_REGEXP)$/) {
        $target = $1;                     # $target now untainted
    } else {
        &$log("Bad data in $target\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }  

    # check $uid and untaint
    if ($uid =~ /^([\d]+)$/) {
        $uid = $1;                     # $uid now untainted
    } else {
        &$log("Bad data in $uid\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }  

    # check $gid and untaint
    if ($gid =~ /^([\d]+)$/) {
        $gid = $1;                     # $gid now untainted
    } else {
        &$log("Bad data in $gid\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }  

    unless ( -f $source) {
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NFL'}";
    }
    if ( -f $target) {
	return "$CONFIG::API{'R_COPY'}:$CONFIG::RESPONSE{'SCS'}";
#	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'FAE'}";
    }
    open SOURCE, "<$source" or return "$CONFIG::API{'CODEDERR'}:$!";
    open TARGET, ">$target" or return "$CONFIG::API{'CODEDERR'}:$!";
    while (<SOURCE>) {
	print TARGET $_;
    }
    close TARGET;
    close SOURCE;
    &$debug("trying to set permissions\n");
    # get permissions of source file
    &$debug("trying to get $source permissions set\n");
    my $perm = &get_file_perm($source);
    &$debug("$source permissions obtained: $perm\n");
    $perm = oct $perm;
    # set permissions on new file
    chmod $perm, $target;
    &$debug("permissions set\n");
    &$debug("trying to chown $target for $uid $gid\n");
    # set ownership of new file
    chown $uid, $gid, $target;
    &$debug("$target chowned\n");
    
    return "$CONFIG::API{'R_COPY'}:1";
}

sub default_link ($$$$) {
    my ($source, $target,$uid,$gid) = @_;
    &$debug("Got $source $target\n");

    # check $target and untaint
    if ($target =~ /^($CONFIG::PATH_REGEXP)$/) {
        $target = $1;                     # $target now untainted
    } else {
        &$log("Bad data in $target\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }  

    # check $uid and untaint
    if ($uid =~ /^([\d]+)$/) {
        $uid = $1;                     # $uid now untainted
    } else {
        &$log("Bad data in $uid\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }  

    # check $gid and untaint
    if ($gid =~ /^([\d]+)$/) {
        $gid = $1;                     # $gid now untainted
    } else {
        &$log("Bad data in $gid\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }  

    # check $source and untaint
    if ($source =~ /^($CONFIG::PATH_REGEXP)$/) {
        $source = $1;                     # $source now untainted
    } else {
        &$log("Bad data in $source\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }  

    unless ( -f $source) {
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NFL'}:$source";
    }

    if ( -f $target) {
	return "$CONFIG::API{'R_LINK'}:1";
    }
    symlink($source, $target) || return "$CONFIG::API{'CODEDERR'}:$!";

    # set ownership of new file
    # note that if the OS does not suport lchown then 
    # we leave the link alone
    if ($CONFIG::OS_HAS_lchown) {
	lchown($uid, $gid, $target);
    }
    return "$CONFIG::API{'R_LINK'}:1";
}


# this routine get a group name  and returns the gid for that group.
sub default_get_gid ($) {
    my $group = $_[0];

    &$debug("Got $group\n");
    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $group\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DGETGID'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'group'} eq $host) {
        &$debug("spawn local $group\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LGETGID'};
        return &$handler_sub(@_);
    #
    } else {
        # 
        $target_host = $CONFIG::ACCT_FILES_MASTER{'group'};
    }
        

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $group\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RGETGID'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# get gid for given group from local DB
sub default_db_get_gid (@) {
    my $group = $_[0];
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth, $gid,$lastchange);

    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);
    
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});
    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }

    # check if group exists
    $sth = $dbh->prepare("SELECT gid,lastchange FROM group_file WHERE gname = \"$group\"") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$gid,\$lastchange);
    $sth->fetch();
    $sth->finish();
    $dbh->disconnect();

    if ($gid) {
	# normalize gid
	$gid--;
	return "$CONFIG::API{'R_GETGID'}:$gid:$lastchange";
    } else {
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSG'}";
    }
}

# Dispatches get gid call to appropriate host
sub default_remote_get_gid (@) {
    return &RatsClientLib::send_remote_command("GETGID",@_);
}

# determine the gid for a given username from local group file
# return uid if succesfull
sub default_local_get_gid ($) {
    my ($gid, $entry, $group_file, @group);
    my ($group, $pass, $members, $target);
    $target = $_[0];

    # read the group file, we don't care is locked
    # maybe we should?
    &$debug("got $target \n");
    ($group, $pass, $gid, $members) = vgetgrnam($target);
    if ($group eq $target) {
	# we add the 0 as placeholder for DB lastchange
	return "$CONFIG::API{'R_GETGID'}:$gid:0";
    } else {
	# read the reserve file
	if (-f $CONFIG::ACCT_FILES{'reserveg'}){
	    my $res_file = $CONFIG::ACCT_FILES{'reserveg'};
	    unless (open(RESERVE, $res_file)) {
		&$log("Failed to open reserve file: $!\n");
		return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'RRE'}";
	    }
	    while (<RESERVE>) {
		chomp;
		($group, $gid)= split(":");
		if ($group eq $target) {
		    return "$CONFIG::API{'R_GETGID'}:$gid:0";
		}
	    }
	}
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSG'}";
    }

}

# this routine get a the groups for a give username
sub default_get_groups ($) {
    my $user = $_[0];

    &$debug("Got $user\n");

    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $user\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DGETGROUP'};
            return &$handler_sub(@_);
        # if remote then set $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    # where is the group file?
    } elsif ($CONFIG::ACCT_FILES_MASTER{'group'} eq $host) {
        &$debug("spawn local $user\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LGETGROUP'};
        return &$handler_sub(@_);
    # if remote file
    } else {
        # set remote host
        $target_host = $CONFIG::ACCT_FILES_MASTER{'group'};
    }
        

    # if we got this far then we send the command to a remote
    &$debug("spawn remote \n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RGETGROUP'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}

# get all groups for a user from the DB
sub default_db_get_groups (@) {
    
    my $user = $_[0];
    
    use DBI;

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth1, $sth2, $group,@groups,$temp,$gid);

    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);
    
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }

    $sth2=$dbh->prepare("SELECT gname FROM group_file WHERE gid = ?") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    

    # first we lookup the primary gid
    # now lookup the primary gid from the password_file
    $sth1=$dbh->prepare("SELECT gid FROM password_file WHERE uname = \"$user\" AND active = '1'") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});

    $sth1->execute() || 
	return &mysql_error($CONFIG::RESPONSE{'DEF'});

    # if no gid for the user the user does not
    # exit or inactive
    ($gid) = $sth1->fetchrow_array();
    unless ($gid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }

    $sth1->finish();
    # look up the name of the primary gid
    $sth2->execute($gid)|| 
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($group) = $sth2->fetchrow_array();
    unless ($group eq "") {
	push @groups, $group;
    }

    $sth1=$dbh->prepare("SELECT gid FROM group_members WHERE uname = \"$user\"") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});

    $sth1->execute() || 
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    
    #for each gid the user has in the group file ...
    while (($gid) = $sth1->fetchrow_array()) {

	# look up the group name
	$sth2->execute($gid) || 
	    return &mysql_error($CONFIG::RESPONSE{'DEF'});
	
	($group) = $sth2->fetchrow_array();
	push @groups, $group;
    }
    $sth1->finish();
    $sth2->finish();


    $dbh->disconnect();
    @groups = &unique(@groups);
    $temp = "";
    $temp = join(":",@groups);
    return "$CONFIG::API{'R_GETGROUP'}:$temp";
}

# Dispatches get groups call to appropriate host
sub default_remote_get_groups (@) {
    return &RatsClientLib::send_remote_command("GETGROUP",@_);
}

# determine the groups for a given username from local group file
# return list if succesfull. Also get the group from the password file
sub default_local_get_groups ($) {
    my $target = $_[0];
    &$debug("got $target\n");
    my (@members, @return_groups, $user, $final_list, $pw_gid);
    my ($gn, $gp, $gid, $mem, $entry, @all_groups, $pwent);
    my $group_file = $CONFIG::ACCT_FILES{'group'};

    # get pw_gid. if we get an error then we return it
    $pwent = &$do('GETPWENT', $target, $CONFIG::ACCT_FILES_MASTER{'passwd'},$CONFIG::CLIENT_PORT);
    my ($code, @data) = split (":", $pwent);
    if ($code == $CONFIG::API{'R_GETPWENT'}) {
	$pw_gid = $data[3];
    } else {
	return $pwent;
    }


    open (GROUP, "$group_file") || return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'GRE'}";
    # find the all the groups that contain the user
    # and the pw_gid
    foreach $entry (<GROUP>) {
	chomp $entry;
	($gn, $gp, $gid, $mem) = split(":", $entry);
	@members = split(",", $mem);
	foreach $user (@members) {
	    if ($user eq $target) {
		push @all_groups, $gn;
	    }
	}
	if ($gid == $pw_gid) {
	    push @all_groups, $gn;
	} 
    }
    close GROUP;

    # reduce them to no ext groups. 
    foreach $entry (@all_groups) {
	($entry) = split ("-$CONFIG::EXT_TOKEN", $entry);
	push @return_groups, $entry;
    }

    @return_groups = &unique(@return_groups);
    $final_list = join (":", @return_groups);
    return "$CONFIG::API{'R_GETGROUP'}:$final_list";
    
}


# This routine generates an uid for a account by obtaining this uid 
# from a remote host for the given user
sub default_generate_sync_uid ($) {
    my $user = $_[0];
    my ($rats_data, $code, @data);
    $rats_data = RatsClientLib::send_remote_command('GETUID',$user,$CONFIG::UID_HOST,$CONFIG::CLIENT_PORT);
    ($code,@data) = split (":", $rats_data);
    unless ($code eq  $CONFIG::API{'R_GETUID'}) {
	return $CONFIG::BOGUS_UID;
    }
    return $data[0];
}


# check for taint
sub is_tainted {
    return  ! eval {
	join ('', @_) , kill 0;
	1;
    };
}

# the is the equivalant of File::Path::rmtree but without the
# taint problems. 
sub vrmtree($) {
    my $fh = new FileHandle;
    my $path = $_[0];
    my $file;
    &$debug("Got $path\n");

    # check path and untaint
    if ($path =~ /^($CONFIG::PATH_REGEXP)$/) {
        $path = $1;                     # $path now untainted
    } else {
        &$log("Bad data in $path\n");        
	return 0;
    }

    if (-l $path) {
	unlink $path || return 0;
    }

    if (-d $path) {
	opendir ($fh, $path) || return 0;
	while ($file = readdir $fh) {
	    unless ($file eq "." || $file eq "..") {
		vrmtree("$path/$file") || return 0;
	    }
	}
	rmdir $path || return 0;
    } else {
	unlink $path || return 0;
    }
    return 1;
}
    

# default archive function, takes a path
sub default_archive ($) {
    my $path = $_[0];
    # figure our where the actual fs is 
    # or if it even exists
    &$debug("Got $path\n");
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");
    my $target_host = $CONFIG::ARCHIVE_HOST;
    if ($host eq $target_host) {
	&$debug("spawn local $path\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'LARCHIVE'};
	return &$handler_sub(@_);
    } else {
	&$debug("spawn remote $path\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'RARCHIVE'};
	return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);
    }
}
# this function takes a path and archives it
sub default_local_archive ($) {
    my $path = $_[0];
    &$debug("Got $path\n");
    my $cmd;
    my $arch_path = $CONFIG::ARCHIVE_PATH;
    # setup some date info
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    $mon++;
    $year = 1900 + $year;
    my $date = sprintf ("%2.2d%2.2d%4.4d", $mon, $mday, $year);


    # check path and untaint
    if ($path =~ /^($CONFIG::PATH_REGEXP)$/) {
        $path = $1;                     # $path now untainted
    } else {
        &$log("Bad data in $path\n");        
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TNT'}";
    }
    
    #make sure archive dir is a dir
    if (-e $arch_path && !-d $arch_path) {
	return "$CONFIG::API{'CODEDERR'}:archive dir not a dir!";
    } 

    # create archive dir if it doesn't exist
    unless (-d $arch_path) {
	mkpath ($arch_path, 0, oct $CONFIG::ARCHIVE_PERMS) || 
	    return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'MKF'}";   
    }

    my $dir = basename $path;

    # if it exists remove it
    if (-d $path) {
	$cmd="$CONFIG::ARCHIVE_CMD $arch_path/$dir.$date.$CONFIG::ARCHIVE_EXT $path";
	system ($cmd)
	    || return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'ARF'}";
    }
    return "$CONFIG::API{'R_ARCHIVE'}:1";
}
    
# Dispatches archive call to appropriate host
sub default_remote_archive ($$$$$) {
    return &RatsClientLib::send_remote_command("ARCHIVE",@_);
}


# return a list of all users on the system
sub default_get_users () {
    &$debug("function entered\n");
    # figure our where the actual password file is maintained

   # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB \n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DGETUSERS'};
            return &$handler_sub(@_);
        # if remote then set $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    # is the password file local?
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
        &$debug("spawn local \n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LGETUSERS'};
        return &$handler_sub(@_);
    # set remote host as password file master
    } else {
        # 
        $target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }
        

    # if we got this far then we send the command to a remote
    &$debug("spawn remote \n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RGETUSERS'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);


}


# get all users from the DB (active or not)
sub default_db_get_users() {

    use DBI;
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth, @users,$temp,$user);
    
    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);
    
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_erro($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    $sth=$dbh->prepare("SELECT uname FROM password_file") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    
    $sth->execute() || 
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    
    while (($user) = $sth->fetchrow_array()) {
	push @users, $user;
    }
    $sth->finish();
    $dbh->disconnect;
    $temp = join(":",@users);
    return "$CONFIG::API{'R_GETUSERS'}:$temp";
}

# Dispatches get users call to appropriate host
sub default_remote_get_users (@) {
    return &RatsClientLib::send_remote_command("GETUSERS",@_);
}

# return a list of all the users in the passwd file
sub default_local_get_users () {
    my @list;
    # read the passwd file, we don't care is locked
    # maybe we should?
    &$debug("entered \n");
    my $pass_file = $CONFIG::ACCT_FILES{'passwd'};
    unless (open(PASSWD, $pass_file)) {
	&$log("Failed to open passwd file: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'PRE'}";
    }
    while (<PASSWD>) {
	chomp;
	my ($usr, $pass, $uid, @rest) = split(":", $_);
	push @list, $usr;
    } 
    close PASSWD;
    my $return = join (":",@list);
    return "$CONFIG::API{'R_GETUSERS'}:$return";
}



# this routine takes a group and return a getgrnam like output with
# API semantics
sub default_getgrnam ($) {
    my ($grp) = $_[0];

    &$debug("Got $grp\n");

   # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $grp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DGETGRNAM'};
            return &$handler_sub(@_);
        # if remote then set $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    # if local file
    } elsif ($CONFIG::ACCT_FILES_MASTER{'group'} eq $host) {
        &$debug("spawn local $grp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LGETGRNAM'};
        return &$handler_sub(@_);
    # remote file
    } else {
        # set remote host
        $target_host = $CONFIG::ACCT_FILES_MASTER{'group'};
    }
        

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $grp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RGETGRNAM'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# do a getgrnam from  a database
sub default_db_getgrnam (@) {
    my $grp = $_[0];

    use DBI;
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth, $gid ,$pass, @members,$member,$temp,$temp2);
    my ($lastchange);
    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);

    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # get the gid and password for group
    $sth = $dbh->prepare("SELECT pass,gid,lastchange FROM group_file WHERE gname = \"$grp\" AND active = \"1\"") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});

    ($pass, $gid, $lastchange) = $sth->fetchrow_array();
    $sth->finish();

    #check if the group exists
    unless ($gid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSG'}";
    }

    #get the group members
    $sth = $dbh->prepare("SELECT uname FROM group_members WHERE gid = $gid") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute()||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    while(($member) = $sth->fetchrow_array()) {
	push @members, $member;
    }
    $sth->finish();

    # build the response
    $temp2= join(",",@members);
    $gid--;
    $temp = "$grp:$pass:$gid:" . $temp2 . ":$lastchange";
    $dbh->disconnect();
    return "$CONFIG::API{'R_GETGRNAM'}:$temp";
	

}



# Dispatches getgrnam call to appropriate host
sub default_remote_getgrnam (@) {
    return &RatsClientLib::send_remote_command("GETGRNAM",@_);
}



# calls the appropriate function to do a getgrnam
sub default_local_getgrnam ($) {
    my $group = $_[0];
    my ($grp, $pass, $task_gid, $members) = vgetgrnam($group);
    my $return = join (":",$grp, $pass, $task_gid, $members);

    # if we didn't find the group then return and cry
    if ($grp eq $CONFIG::RESPONSE{'GRE'}) {
	return "$CONFIG::API{'CODEDERR'}:$grp";
    }
    
    # we add the the 0 as placeholder for DB timestamp
    return "$CONFIG::API{'R_GETGRNAM'}:$return:0";
}




# this routine get a username and returns the uid for that user.
# if the username does not exist it allocates one and saves it 
# in a reserved uid file. This function decides where the action 
# should take place and spawn the appropriate functions
sub default_reserve_uid ($) {
    my $usr = $_[0];

    &$debug("Got $usr\n");

    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DRESERVE'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
        &$debug("spawn local $usr\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LRESERVE'};
        return &$handler_sub(@_);
    #
    } else {
        # 
        $target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }
        

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RRESERVE'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}

# reserves uid in local DB by adding username and uid only with 
# inactive flag in the password_file table
sub default_db_reserve_uid (@) {
    my $usr = $_[0];
    use DBI;

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth, @pass_parts);
    my ($uname, $pass, $uid, $gcos, $home, $shell, $last, $min, $max, $warn);
    my ($inactive, $expire, $flag, $exp, $flag, $chg, $active, $gid);

    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);
    
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});
    
    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # check if user already exists
    $sth = $dbh->prepare("SELECT uid FROM password_file WHERE uname = \"$usr\"") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid);
    $sth->fetch();
    $sth->finish();
    if ($uid) {
	$uid--;
	$dbh->disconnect();
	return "$CONFIG::API{'R_RESERVE'}:$uid";
    }

    # if user does not already exits then we insert him;
    $dbh->do("INSERT password_file (uname,uid,active) VALUES(\"$usr\", NULL,\"0\")") ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    # obtain the new UID
    $uid = $dbh->{'mysql_insertid'};

    # normalize it
    $uid--;

    #clean up
    $dbh->disconnect();
    return "$CONFIG::API{'R_RESERVE'}:$uid";
    
    
}


# Dispatches reserve uid call to appropriate host
sub default_remote_reserve_uid (@) {
    return &RatsClientLib::send_remote_command("RESERVE",@_);
}

# create a reserved uid
# note that it assumes that it running on the password host
# and it will do some lookups on the local host directly

sub default_local_reserve_uid ($) {
    my ($uid, $entry, $res_file, $res_temp, $pass_file);
    my ($target, $result, $code, @data, $handler_sub);
    $target = $_[0];

    # check if the person already has an uid
    $handler_sub = $CONFIG::COMMAND_HOOKS{'LGETUID'};
    $result  =  &$handler_sub($target);
    ($code, @data) = split (":", $result);
    if ($code eq $CONFIG::API{'R_GETUID'}) {
	return "$CONFIG::API{'R_RESERVE'}:$data[0]";
    }

    # if we don't allow reserverasions that's tough.
    unless ($CONFIG::ALLOW_UID_RESERVE) {
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'RNS'}";
    }


    #lock the password file
    &$debug("locking password\n");
    $pass_file = $CONFIG::ACCT_FILES{'passwd'};
    $result = &lock_file('passwd');
    unless ($result eq $CONFIG::RESPONSE{'SCS'}) {
	&$log("Failed passwd lock: $result!\n");
	return "$CONFIG::API{'CODEDERR'}:$result";
    }

    #lock the reserve file
    &$debug("locking reserve\n");
    $pass_file = $CONFIG::ACCT_FILES{'reserve'};
    $result = &lock_file('reserve');
    unless ($result eq $CONFIG::RESPONSE{'SCS'}) {
	&unlock_file('passwd');
	&unlock_file('reserve');
	&$log("Failed reserve lock: $result!\n");
	return "$CONFIG::API{'CODEDERR'}:$result";
    }


    #generate new uid
    $handler_sub = $CONFIG::COMMAND_HOOKS{'GENNUID'};
    $uid = &$handler_sub($target);
    # check if it failed
    if ($uid == $CONFIG::BOGUS_UID) {
	&unlock_file('reserve');
	&unlock_file('passwd');
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'UGF'}";
    }

    # unlock all files
    &unlock_file('passwd');
    &unlock_file('reserve');

    &$debug("returning $uid\n");
    return "$CONFIG::API{'R_RESERVE'}:$uid";
}


# the default generate uid function return the next highes avaible uid
# based on the current password file. 
#
# WARNING: this function assumes that you already have the password file
# WARNING: locked. Otherwise multiple proccess could get the same uid 
# WARNING: before the first one finished updating the file
# 
# Note that it also checks the reserved file to make sure that it catches 
# all the pre allocated UID's
#
# it returns the uid or BOGUS_UID for failure
sub default_generate_next_uid  ($) {
    my $arg = $_[0];
    my ($max_uid, $uid, $entry, $pass_file, @passwd);
    my ($usr, $pass, @rest, $res_file, $res_temp);
    my (@reserve, %uid_space);

    # the max uid (which is what we are going to add 1 to and return
    # can't be less the MIN_UID - 1 
    $max_uid = $CONFIG::MIN_UID - 1;

    # read the passwd file
    $pass_file = $CONFIG::ACCT_FILES{'passwd'};
    unless (open(PASSWD, $pass_file)) {
	&$log("Failed to open passwd file: $!\n");
	return $CONFIG::BOGUS_UID;
    }
    @passwd = <PASSWD>;
    close PASSWD;

    # read the reserve file
    if ($CONFIG::ACCT_FILES{'reserve'} && -f $CONFIG::ACCT_FILES{'reserve'}) {
	$res_file = $CONFIG::ACCT_FILES{'reserve'};
	unless (open(RESERVE, $res_file)) {
	    &$log("Failed to open reserve file: $!\n");
	    return $CONFIG::BOGUS_UID;
	}
	@reserve = <RESERVE>;
	close RESERVE;
    }


    if ($CONFIG::USE_NEXT_UID) { # get the next uid
	&$debug("looking for first free\n");
	# find the used uid's
	foreach $entry (@passwd) {
	    ($usr, $pass, $uid, @rest) = split(":", $entry);
	    $uid_space{$uid} = 1;
	}
	foreach $entry (@reserve) {
	    ($usr, $uid) = split(":", $entry);
	    $uid_space{$uid} = 1;
	}
	for ($max_uid = $CONFIG::MIN_UID;;$max_uid++) {
	       last unless $uid_space{$max_uid};
	}

    } else {  	# find the max uid
	&$debug("looking for largest free\n");
	foreach $entry (@passwd) {
	    ($usr, $pass, $uid, @rest) = split(":", $entry);
	    if ($uid > $max_uid) {
		$max_uid = $uid;
	    }
	}
	
	#check for max in reserved file
	# find the max uid
	foreach $entry (@reserve) {
	    chomp $entry;
	    ($usr, $uid) = split(":", $entry);
	    if ($uid > $max_uid) {
		$max_uid = $uid;
	    }
	}
	# generate the next uid
	$max_uid++;
    }
    &$debug("next uid is $max_uid\n");
    # update reserve file
    $res_temp = $CONFIG::ACCT_FILES_TMP{'reserve'};
    unless (open(TEMP, ">$res_temp")) {
	&$log("Failed reserve open: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'RTE'}";
    }
    
    $res_file = $CONFIG::ACCT_FILES{'reserve'};
    if (-f $res_file) {   
	unless (open(RESERVE, $res_file)) {
	    &$log("Failed reserve open: $!\n");
	    return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'RRE'}";
	}
	# copy existsing data
	while (<RESERVE>) { print TEMP;}
    }

    # put in new entry
    print TEMP "$arg:$max_uid\n";

    # close all file
    close RESERVE; close TEMP;

    # move reserve file into place
    tmove ($res_temp, $res_file);
    chmod oct $CONFIG::ACCT_FILES_PERMS{'reserve'}, $res_file;
    chown $CONFIG::ACCT_FILES_OWNER{'reserve'}, $CONFIG::ACCT_FILES_GROUP{'reserve'}, $res_file;

    # return the next highest uid
    return $max_uid;
}




# this routine takes a filepath as an arg and returns the
# permissions of that file.
sub get_file_perm ($) {
    use File::stat;
    my $file = $_[0];
    my ($st, $mod);
    $mod = "0000";
    if ( -f $file ) {
	$st = stat $file;
	$mod = $st->mode;    
	# get the file mode. A tad of magic
	# read the stat.h files for kicks
	my $bits = ($mod & 0x1ff); # get permissions
	my $bits1 = ($mod & 0x800); # setuid?
	my $bits2 = ($mod & 0x400); # setgid?
	$bits = $bits + $bits1 + $bits2; # all permissions?
	$mod = sprintf "%04o", $bits;
    }
    &$debug("get_file_perm: returning $mod\n");
    return $mod;
}


# this routine implements the lchown call provided by Solaris.
# this may or may not work under other os'es if the $CONFIG::OS_HAS_lchown
# and $CONFIG::SYS_lchown have been well set.
# NOTE that the arguments are taken in the same order as those of the
# perl chown even though lchown(2) and chown(2) have a different order.
sub lchown ($$$) {
    my ( $uid, $gid, $file )= @_;
    if ($CONFIG::OS_HAS_lchown) {
        return syscall($CONFIG::SYS_lchown, $file, $uid+0, $gid+0);
    } else {
        return -1;
    }
}


# get pwent. takes username returns shadow entry or error
sub default_getshent($) {
    my $usr = $_[0];
    &$debug("Got $usr\n");
    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DGETSHENT'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    # where is the shadow file
    } elsif ($CONFIG::ACCT_FILES_MASTER{'shadow'} eq $host) {
        &$debug("spawn local $usr\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LGETSHENT'};
        return &$handler_sub(@_);
    #
    } else {
        # set remote host
        $target_host = $CONFIG::ACCT_FILES_MASTER{'shadow'};
    }
        

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RGETSHENT'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);


}

# get sh ent from local db. returns
# shadow entry or error.
sub default_db_getshent (@) {
    my $user = $_[0];
    use DBI;
    
    &$debug("got $user\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth, @pass_parts,$temp);
    my ($uname, $pass, $uid, $gcos, $home, $shell, $last, $min, $max, $warn);
    my ($inactive, $expire, $flag, $exp, $flag, $chg, $active, $gid);

    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);

    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # there should be a max of one row, because usernames are unique
    $sth = $dbh->prepare("SELECT * FROM password_file WHERE uname = \"$user\" AND active = \"1\"") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});


    $sth->bind_columns (\$uname, \$pass, \$uid, \$gid,\$gcos, \$home,
                    \$shell, \$last, \$min, \$max, \$warn,
                    \$inactive, \$exp, \$flag, \$chg, \$active);

    # there should be a max of one row, because usernames are unique
    $sth->fetch();
    
    $sth->finish();
    $dbh->disconnect();

    if ($uname) {
	$temp = "$uname:$pass:$last:$min:$max:$warn:$inactive:$exp:$flag:$chg";
	return "$CONFIG::API{'R_GETSHENT'}:$temp";
    } else {
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
}

# Dispatches getshent call to appropriate host
sub default_remote_getshent (@) {
    return &RatsClientLib::send_remote_command("GETSHENT",@_);
}

# get shent. takes username returns shadow entry or error
sub default_local_getshent ($) {
    my ($uid, $entry, $shadow_file, @shadow);
    my ($usr, $pass, @rest, $target, $val);
    $target = $_[0];

    $val = "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    # read the shadow file, we don't care is locked
    # maybe we should?
    &$debug("got $target \n");
    $shadow_file = $CONFIG::ACCT_FILES{'shadow'};
    unless (open(SHADOW, $shadow_file)) {
	&$log("Failed to open shadow file: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'SRE'}";
    }
    while (<SHADOW>) {
	chomp;
	($usr, @rest) = split(":", $_);
	if ($target eq $usr) {
	    # adding 0 as placeholder for DB timestamp
	    $val = "$CONFIG::API{'R_GETSHENT'}:$_:0";
	} 
    }
    close SHADOW;

    &$debug("returning $val\n");
    return $val;
}


# check if user exists. takes username returns 0 or 1
sub default_exist_user($) {
    my $usr = $_[0];
    &$debug("Got $usr\n");
    # figure out where we are
    my $host = &get_hostname();
    &$debug("Determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DEXUSER'};
            return &$handler_sub(@_);
        # if remote then set $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    # where is the password file
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
        &$debug("spawn local $usr\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LEXUSER'};
        return &$handler_sub(@_);
    #
    } else {
        # set remote host
        $target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }
        

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'REXUSER'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# check if user exists. takes username returns 0 or 1.
# it check both active and inactive accounts in local DB
sub default_db_exist_user (@) {
    my $usr = $_[0];
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth, $uid);

    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);
    
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});
    
    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # check if user exists
    $sth = $dbh->prepare("SELECT uid FROM password_file WHERE uname = \"$usr\"") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid);
    $sth->fetch();
    $sth->finish();
    $dbh->disconnect();

    # if we got an uid then user exists else you loose.
    if ($uid) {
	return "$CONFIG::API{'R_EXUSER'}:1";
    } else {
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
}

# Dispatches exist_user call to appropriate host
sub default_remote_exist_user (@) {
    return &RatsClientLib::send_remote_command("EXUSER",@_);
}

# check if user exists. takes username returns 0 or 1.
# it check both passwd and reserve
sub default_local_exist_user ($) {
    my ($uid, $entry, $passwd_file);
    my ($usr, $pass, @rest, $target, $val);
    my ($reserve_file, $found);
    $target = $_[0];
    &$debug("got $target \n");

    # read the passwd file, we don't care is locked
    # maybe we should?
    $passwd_file = $CONFIG::ACCT_FILES{'passwd'};
    unless (open(PASSWD, $passwd_file)) {
	&$log("Failed to open passwd file: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'PRE'}";
    }
    while (<PASSWD>) {
	chomp;
	($usr, @rest) = split(":", $_);
	if ($target eq $usr) {
	    &$debug("returning 1 for $target\n");
	    return "$CONFIG::API{'R_EXUSER'}:1";
	} 
    }
    close PASSWD;


    # read the reserve file, we don't care is locked
    # maybe we should?
    $reserve_file = $CONFIG::ACCT_FILES{'reserve'};
    if ( -f $reserve_file) {
	unless (open(RESERVE, $reserve_file)) {
	    &$log("Failed to open passwd file: $!\n");
	    return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'RRE'}";
	}
	while (<RESERVE>) {
	    chomp;
	    ($usr, @rest) = split(":", $_);
	    if ($target eq $usr) {
		&$debug("returning 1 for $target\n");
		return "$CONFIG::API{'R_EXUSER'}:1";
	    } 
	}
	close RESERVE;
    }

    &$debug("returning NSU for $target\n");
    return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
}

sub mysql_error($$) {

    my ($caller, $text) = @_;
    my $host = &get_hostname();
    my $name = basename $0;

    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    # $mon is reported in 0->11 so we add 1 for human consumption
    $mon++;
    $year += 1900;
    my $calling = (caller(1))[3];
    my $msg = "$mon/$mday/$year $hour:$min:$sec $host:$name:$$:$calling: $text";

    &$log("REMOTE", "$msg: $text $DBI::err $DBI::errstr\n");
    return "$CONFIG::API{'CODEDERR'}:$text:$DBI::err:$DBI::errstr";
}



# gets group name pass and gid and dispatches to appropriate 
# host/db/file for addition
sub default_add_grp_ent(@)
{
    my ($group_name,$group_pass,$group_id) = @_;
    &$debug("Got $group_name,$group_pass,$group_id\n");

    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
	# if local then do it
	if ($CONFIG::CLIENT_DB_HOST eq $host) {
	    &$debug("spawn DB $group_name,$group_pass,$group_id\n");
	    my $handler_sub = $CONFIG::COMMAND_HOOKS{'DADDGRPENT'};
	    return &$handler_sub(@_);
	# if remote then sety $target_host
	} else {
	    $target_host = $CONFIG::CLIENT_DB_HOST;
	}
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'group'} eq $host) {
	&$debug("spawn local $group_name,$group_pass,$group_id\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'LADDGRPENT'};
	return &$handler_sub(@_);
    #
    } else {
	# 
	$target_host = $CONFIG::ACCT_FILES_MASTER{'group'};
    }
	

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $group_name,$group_pass,$group_id\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RADDGRPENT'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);
}



# this function takes group entry parts, merges them, and adds
# the to a group file
# it returns SCS on success or the appropriate error message otherwise
sub default_local_add_grp_ent (@) {

    # arg!! the number of args!
    my ($group,$pass,$gid) = @_;
    my (@group, $entry, $grp_file, $temp_res, $temp_entry);
    my ($temp_grp, @rest, $handler_sub, $grp_file, @reserve);
    my ($response, $temp_gid, $res_file, $temp_pass, $already_reserved);

    # now we lock various files
    $response = &lock_file('group');
    unless ($response eq $CONFIG::RESPONSE{'SCS'}) {
	&$log("Failed group lock: $response!\n");
	return "$CONFIG::API{'CODEDERR'}:$response";
    }
    #lock the reserve file
    $res_file = $CONFIG::ACCT_FILES{'reserveg'};
    $response = &lock_file('reserveg');
    unless ($response eq $CONFIG::RESPONSE{'SCS'}) {
	&unlock_file('reserveg');
	&unlock_file('group');
	&$log("Failed reserve lock: $response!\n");
	return "$CONFIG::API{'CODEDERR'}:$response";
    }

    # now we read the whole group file into memory
    # this is sort of wastfull. Maybe we should just run through it
    $grp_file = $CONFIG::ACCT_FILES{'group'};
    unless (open(GROUP, "$grp_file")) {
	&unlock_file('group');
	&unlock_file('reserveg');
	&$log("Failed group open: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'GRE'}";
    }
    @group = <GROUP>;
    close GROUP;


    # now we read the whole reserve file into memory
    $res_file = $CONFIG::ACCT_FILES{'reserveg'};
    if (-f $res_file) {
	unless (open(RES, "$res_file")) {
	    &unlock_file('group');
	    &unlock_file('reserveg');
	    &$log("Failed reserve open: $!\n");
	    return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'RRE'}";
	}
	@reserve = <RES>;
	close RES;
    }

    # first we verify that the group does not already exist
    # or the gid was already used
    foreach $entry (@group) {
	($temp_grp, $temp_pass, $temp_gid, @rest) = split(":", $entry);
	if (($temp_grp eq $group) || ($temp_gid == $gid)) {
	    &unlock_file('group');
	    &unlock_file('reserveg');
	    &$log("$group or gid $gid already exists\n");
	    return "$CONFIG::API{'ITEXISTS'}:$CONFIG::RESPONSE{'GAE'}";
	}
    }

    # check the reserve file for gid 
    foreach $entry (@reserve) {
	$temp_entry= $entry;
	chomp $temp_entry;
	($temp_grp, $temp_gid) = split(":", $temp_entry);
	if (($temp_grp eq $group) && ($temp_gid != $gid) && ($gid != $CONFIG::BOGUS_GID)) {
	    &unlock_file('group');
	    &unlock_file('reserveg');
	    &$log("$group or gid $gid already exists\n");
	    return "$CONFIG::API{'ITEXISTS'}:$CONFIG::RESPONSE{'GAE'}";
	} elsif (($temp_grp eq $group) && ($gid == $CONFIG::BOGUS_GID)) {
	    $already_reserved = 1;
	    $gid = $temp_gid;
	} elsif (($temp_grp ne $group) && ($gid == $temp_gid)) {
	    &unlock_file('group');
	    &unlock_file('reserveg');
	    &$log("$group or gid $gid already exists\n");
	    return "$CONFIG::API{'ITEXISTS'}:$CONFIG::RESPONSE{'GAE'}";
	}
    }

    # now that we have the the file locked we can generate a GID.
    # if the gid is the bogus (according to config file) then we 
    # generate one. Otherwise we use the  given one
    if ($gid == $CONFIG::BOGUS_GID) {
	$handler_sub = $CONFIG::COMMAND_HOOKS{'GENGID'};
	$gid = &$handler_sub($group);
	# check if it failed
	if ($gid == -1) {
	    &unlock_file('group');
	    &unlock_file('reserveg');
	    return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'GGF'}";
	}
	$already_reserved = 1;
    }

    # now we can generate the group entry
    $entry = join(":", $group,$pass,$gid) . ":";
    
    # write out the new group file to a temp file
    $temp_grp = $CONFIG::ACCT_FILES_TMP{'group'};
    unless (open(GR_TEMP, ">$temp_grp")) {
	&unlock_file('group');
	&unlock_file('reserveg');
	&$log("Failed group temp open: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'GTE'}";
    }
    print GR_TEMP @group;
    print GR_TEMP "$entry\n";
    close GR_TEMP;
    

        #update reserve file if not already
    unless ($already_reserved) {
	# write out the new reserve file to a temp file
	$temp_res = $CONFIG::ACCT_FILES_TMP{'reserveg'};
	unless (open(RS_TEMP, ">$temp_res")) {
	    &unlock_file('group');
	    &unlock_file('reserveg');
	    &$log("Failed reserve temp open: $!\n");
	    return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'RTE'}";
	}
	print RS_TEMP @reserve;
	print RS_TEMP "$group:$gid\n";
	close RS_TEMP;
	tmove ($temp_res, $res_file);
	chmod oct $CONFIG::ACCT_FILES_PERMS{'reserveg'}, $res_file;
	chown $CONFIG::ACCT_FILES_OWNER{'reserveg'}, $CONFIG::ACCT_FILES_GROUP{'reserveg'}, $res_file;
    }



    # and we move them in the right place
    tmove ($temp_grp, $grp_file);
    chmod oct $CONFIG::ACCT_FILES_PERMS{'group'}, $grp_file;
    chown $CONFIG::ACCT_FILES_OWNER{'group'}, $CONFIG::ACCT_FILES_GROUP{'group'}, $grp_file;

    # and unlock both
    &unlock_file('group');
    &unlock_file('reserveg');
    &$debug("$group added to group file\n");
    return "$CONFIG::API{'R_ADDGRPENT'}:$gid";
}

# perform remote group addition 
# gets gname/pass/gid/host/port returns gid or error
sub default_remote_add_grp_ent(@)
{
    return &RatsClientLib::send_remote_command("ADDGRPENT",@_);
}

# perform local db group addition 
# gets gname/pass/gid returns gid or error
sub default_db_add_grp_ent(@)
{
    my ($group_name,$group_pass,$gid_ext) = @_;

    &$debug("got $group_name,$group_pass,$gid_ext\n");

    my ($gid_int,$old_gname,$old_gid_int,$old_active);
    use DBI;
    my ($group_active);
	
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sql, @pass_parts,$timeout,$lock);
    
    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);

    $gid_int = $gid_ext+1; #add 1 to external passed gids and uids always match
                          #flat file
    
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }


    # lock entry while we muck around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sql = $dbh->prepare("SELECT GET_LOCK(\"group_file-$group_name\",\"$timeout\")") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});

    $sql->execute() ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sql->fetchrow_array();
    if ($lock == 0) {
	$sql->finish();
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
	$sql->finish();
	$dbh->disconnect();
    	return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }

    $sql = $dbh->prepare("SELECT gname, gid, active FROM group_file where gname = ? OR gid = ?")
	|| return &mysql_error($CONFIG::RESPONSE{'DPF'});
    
    $sql->execute($group_name, $gid_int) ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    
    $sql->bind_columns (\$old_gname,\$old_gid_int,\$old_active);

    while($sql->fetch())
    {
	#see if the gname is taken
	if($old_gname eq $group_name && $old_active == "1")
	{
	    $sql->finish();
	    $dbh->disconnect();
	    return "$CONFIG::API{'CODEDERR'}:Cannot reuse active gname";
	}
	#see if the uid is taken by another group
	if ($old_gid_int == $gid_int && $old_gname ne $group_name)
	{
	    $sql->finish();
	    $dbh->disconnect();
	    return "$CONFIG::API{'CODEDERR'}:GID is in use.";
	}
	#see if we need to generate a new UID
	if ($old_gname eq $group_name && $old_gid_int != $gid_int && $gid_int != $CONFIG::BOGUS_GID + 1)
	{
	    $sql->finish();
	    $dbh->disconnect();
	    return "$CONFIG::API{'CODEDERR'}:Existing inactive and new GID don't match.";
	}
	#if gname matches, uid doesn't, but got autogen value, reuse old GID
	if ($old_gname eq $group_name && $gid_int == $CONFIG::BOGUS_UID + 1) 
	{
	    $gid_int = $old_gid_int;
	    $gid_ext = $old_gid_int-1;
	}
    }
    $sql->finish();

    #insert new entry using replace, shoudl be okay if we got this far
    $sql = $dbh->prepare(qq{REPLACE group_file VALUES(?,?,?,?,?)})||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});

    $sql->execute("$group_name","$group_pass","$gid_int",undef,"1")
	|| return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sql->finish();
    
    #check uid in case we did autoincrement
    unless ($gid_int) {
	$gid_int = $dbh->{'mysql_insertid'};
    }

    #fix_ext to match
    $gid_ext = $gid_int-1;
    
    #closedb
    $dbh->disconnect();

    #return success
    return "$CONFIG::API{'R_ADDGRPENT'}:$gid_ext";
    

}



# gets group name and dispatches to appropriate 
# host/db/file for remocal
sub default_rm_grp_ent(@)
{
    my ($group_name) = @_;
    &$debug("Got $group_name\n");

    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
	# if local then do it
	if ($CONFIG::CLIENT_DB_HOST eq $host) {
	    &$debug("spawn DB $group_name");
	    my $handler_sub = $CONFIG::COMMAND_HOOKS{'DREMGRPENT'};
	    return &$handler_sub(@_);
	# if remote then sety $target_host
	} else {
	    $target_host = $CONFIG::CLIENT_DB_HOST;
	}
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'group'} eq $host) {
	&$debug("spawn local $group_name\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'LREMGRPENT'};
	return &$handler_sub(@_);
    #
    } else {
	# 
	$target_host = $CONFIG::ACCT_FILES_MASTER{'group'};
    }
	

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $group_name\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RREMGRPENT'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);
}




# this function takes a username and remove the passwd/shadow entries for it
# it returns SCS on success or the appropriate error message otherwise
sub default_local_rm_grp_ent (@) {

    my $grp = $_[0];
    my (@group, $entry, $grp_file);
    my ($temp_grp, @rest, $temp_group);
    my ($response, $found);

    # now we lock various files
    $response = &lock_file('group');
    unless ($response eq $CONFIG::RESPONSE{'SCS'}) {
	&$log("Failed group lock: $response!\n");
	return "$CONFIG::API{'CODEDERR'}:$response";
    }

    # now we read the whole group file into memory
    $grp_file = $CONFIG::ACCT_FILES{'group'};
    unless (open(GROUP, "$grp_file")) {
	&unlock_file('group');
	&$log("Failed group open: $!\n");
	return"$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'GRE'}";
    }
    @group = <GROUP>;
    close GROUP;

    # first we verify that the group exist
    $found = 0;
    foreach $entry (@group) {
	($temp_grp, @rest) = split(":", $entry);
	if ($temp_grp eq $grp) {
	    $found = 1;
	}
    }
    unless ($found) {
	&unlock_file('group');
	&$log("$grp does not exist\n");
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSG'}";
    }


    # write out the new group file to a temp file
    $temp_group = $CONFIG::ACCT_FILES_TMP{'group'};
    unless (open(GR_TEMP, ">$temp_group")) {
	&unlock_file('group');
	&$log("Failed group temp open: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'GTE'}";
    }
    foreach $entry (@group) {
	($temp_grp, @rest) = split(":", $entry);
	print GR_TEMP $entry unless ($temp_grp eq $grp);
    }
    close GR_TEMP;

    # and we move it in the right place
    tmove ($temp_group, $grp_file);
    chmod oct $CONFIG::ACCT_FILES_PERMS{'group'}, $grp_file;
    chown $CONFIG::ACCT_FILES_OWNER{'group'}, $CONFIG::ACCT_FILES_GROUP{'group'}, $grp_file;

    # and unlock 
    &unlock_file('group');
    &$debug("$grp removed from group file\n");
    return "$CONFIG::API{'R_REMGRPENT'}:1";
}

# perform remote group removal 
# gets gname  returns 1 or error
sub default_remote_rm_grp_ent(@)
{
    return &RatsClientLib::send_remote_command("REMGRPENT",@_);
}

# perform local db group removal
# it first removes all the group members and then the group itself.
# gets gname 1 or error
sub default_db_rm_grp_ent(@)
{
    my ($group_name) = @_;

    &$debug("got $group_name\n");

    my ($gid);
    use DBI;

	
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sql, @pass_parts, $timeout, $lock);
    
    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);

    
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});
    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }


    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sql = $dbh->prepare("SELECT GET_LOCK(\"group_file-$group_name\",\"$timeout\")") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});

    $sql->execute() ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sql->fetchrow_array();
    if ($lock == 0) {
	$sql->finish();
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
	$sql->finish();
	$dbh->disconnect();
    	return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }


    $sql = $dbh->prepare("SELECT gid FROM group_file where gname = ?")
	|| return &mysql_error($CONFIG::RESPONSE{'DCF'});
    
    $sql->execute($group_name) ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    
    $sql->bind_columns (\$gid);
    $sql->fetch();
    $sql->finish();

    # if no such group say so.
    unless ($gid) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'NSG'}";
    }


    # mark group as inactive
    $sql = $dbh->do("UPDATE group_file SET active = \"0\" WHERE gname = \"$group_name\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    $dbh->disconnect();
    return "$CONFIG::API{'R_REMGRPENT'}:1";

}



sub default_reserve_gid(@)
{
    my ($gname) = @_;
    &$debug("Got $gname\n");
    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
	# if local then do it
	if ($CONFIG::CLIENT_DB_HOST eq $host) {
	    &$debug("spawn DB $gname\n");
	    my $handler_sub = $CONFIG::COMMAND_HOOKS{'DRESERVEG'};
	    return &$handler_sub(@_);
	# if remote then sety $target_host
	} else {
	    $target_host = $CONFIG::CLIENT_DB_HOST;
	}
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'group'} eq $host) {
        &$debug("spawn local $gname\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LRESERVEG'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'group'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $gname");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RRESERVEG'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);
}


# create a reserved gid
# note that it assumes that it running on the group host
# and it will do some lookups on the local host directly

sub default_local_reserve_gid ($) {
    my ($gid, $entry, $res_file, $res_temp, $grp_file);
    my ($target, $result, $code, @data, $handler_sub);
    $target = $_[0];

    # check if the group already has an gid
    $handler_sub = $CONFIG::COMMAND_HOOKS{'LGETGID'};
    $result  =  &$handler_sub($target);
    ($code, @data) = split (":", $result);
    if ($code eq $CONFIG::API{'R_GETGID'}) {
	return "$CONFIG::API{'R_RESERVEG'}:$data[0]";
    }

    # if we don't allow reserverasions that's tough.
    unless ($CONFIG::ALLOW_GID_RESERVE) {
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'RNS'}";
    }


    #lock the group file
    &$debug("locking group\n");
    $grp_file = $CONFIG::ACCT_FILES{'group'};
    $result = &lock_file('group');
    unless ($result eq $CONFIG::RESPONSE{'SCS'}) {
	&$log("Failed group lock: $result!\n");
	return "$CONFIG::API{'CODEDERR'}:$result";
    }

    #lock the reserve file
    &$debug("locking reserve\n");
    $res_file = $CONFIG::ACCT_FILES{'reserveg'};
    $result = &lock_file('reserveg');
    unless ($result eq $CONFIG::RESPONSE{'SCS'}) {
	&unlock_file('group');
	&unlock_file('reserveg');
	&$log("Failed reserve lock: $result!\n");
	return "$CONFIG::API{'CODEDERR'}:$result";
    }


    #generate new gid
    $handler_sub = $CONFIG::COMMAND_HOOKS{'GENNGID'};
    $gid = &$handler_sub($target);
    # check if it failed
    if ($gid == $CONFIG::BOGUS_GID) {
	&unlock_file('reserveg');
	&unlock_file('group');
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'UGF'}";
    }

    # unlock all files
    &unlock_file('group');
    &unlock_file('reserveg');

    &$debug("returning $gid\n");
    return "$CONFIG::API{'R_RESERVEG'}:$gid";
}


sub default_remote_reserve_gid(@)
{
    return &RatsClientLib::send_remote_command("RESERVEG",@_);
}

sub default_db_reserve_gid(@)
{
    my ($gname) = @_;

    &$debug("Got $gname\n");
   
    my($gid);
    


    use DBI;
    my $entryexists =0;
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sql);
    
    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);
    
    my $tryloop=0;
    my $count=-1;
   
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
   
    #see if group already exists
    $sql = $dbh->prepare("SELECT gid FROM group_file WHERE gname = \"$gname\"")
	 ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sql->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sql->bind_columns (\$gid);
    $sql->fetch();
    $sql->finish();

    #if exists return the allocated GID
    if ($gid) {
        $gid--;
        $dbh->disconnect();
        return "$CONFIG::API{'R_RESERVEG'}:$gid";
    }

    #if they don't exist, insert/reserve GID and return generated GID
    $dbh->do("INSERT group_file (gname,gid,active) VALUES(\"$gname\", NULL,\"0\"
)") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    # obtain the new UID
    $gid = $dbh->{'mysql_insertid'};

    # normalize it
    $gid--;

    #clean up
    $dbh->disconnect();
    return "$CONFIG::API{'R_RESERVEG'}:$gid";

}



# default change shell call. It figures out
# where the passwd file resides and dispatches the call
sub default_change_shell ($$$) {
    my ($usr,$shell,$timestamp) = @_;

 
    &$debug("Got $usr $shell $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $shell $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGSHELL'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
        &$debug("spawn local $usr $shell $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGSHELL'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $shell $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGSHELL'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change shell call to appropriate host
sub default_remote_change_shell (@) {
    return &RatsClientLib::send_remote_command("CHGSHELL",@_);
}

# changes the shell of a user by editing passwd file
sub default_local_change_shell ($$$) {
    my ($target_user, $new_shell, $timestamp) = @_;
    &$debug("got $target_user, $new_shell, $timestamp\n");
    return default_change_local_passwd_file($target_user,'shell',$new_shell,$CONFIG::API{'R_CHGSHELL'});
}

# reserves uid in local DB by adding username and uid only with
# inactive flag in the password_file table
sub default_db_change_shell (@) {
    my ($target_user,$new_shell,$timestamp) = @_;
    my($lock,$timeout,$uid,$local_timestamp);
    use DBI;
    &$debug("got $target_user, $new_shell $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});
    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }

    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$target_user\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT uid, lastchange FROM password_file WHERE uname = \"$target_user\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($uid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE password_file set shell = \"$new_shell\", lastchange = NULL where uname = \"$target_user\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up

    # get the timestamp
    my $newt = &get_passlast($dbh, $target_user);
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGSHELL'}:$newt";
}



# add an account type to a user.
# takes username and account type
sub default_add_acct ($$) {
    my ($user, $acct) = @_;
    &$debug("Got $user, $acct");
    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $user, $acct\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DADDACCT'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
        &$debug("spawn local $user, $acct\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LADDACCT'};
        return &$handler_sub(@_);
    #
    } else {
        # 
        $target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }
        

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $user, $acct\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RADDACCT'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);
}

# send remote add acct call to the appropriate host
sub default_remote_add_acct ($$$$) { 
    return &RatsClientLib::send_remote_command("ADDACCT",@_);
}

# add account type to local file
sub default_local_add_acct ($$) {
    my ($user,$acct) = @_;
    &$debug("Got $user\n");
    my $found = 0;
    my $acc_file = $CONFIG::ACCT_FILES{'accounts'};
    my $acc_temp = $CONFIG::ACCT_FILES_TMP{'accounts'};

    #lock the accounts file
    &$debug("locking accounts\n");
    my $result = &lock_file('accounts');
    unless ($result eq $CONFIG::RESPONSE{'SCS'}) {
	&unlock_file('accounts');
	&$log("Failed accounts lock: $result!\n");
	return "$CONFIG::API{'CODEDERR'}:$result";
    }

    #open temp file
    unless (open(TEMP, ">$acc_temp")) {
	&unlock_file('accounts');
	&$log("Failed accounts open: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'ATE'}";
    }

    #copy existing data if any
    if (-f $acc_file) {   
	unless (open(ACCT, $acc_file)) {
	    &$log("Failed accounts open: $!\n");
	    return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'ARE'}";
	}
	# copy existsing data
	while (<ACCT>) { 
	    print TEMP;
	    chomp;
	    my($oldu, $olda) =  split(":");
	    if ($oldu eq $user && $olda eq $acct) {
		$found = 1;
	    }
	}
    }

    # put in new entry
    unless ($found) {
	print TEMP "$user:$acct\n";
    }

    # close all file
    close ACCT; close TEMP;

    # move reserve file into place
    tmove ($acc_temp, $acc_file);
    chmod oct $CONFIG::ACCT_FILES_PERMS{'accounts'}, $acc_file;
    chown $CONFIG::ACCT_FILES_OWNER{'accounts'}, $CONFIG::ACCT_FILES_GROUP{'accounts'}, $acc_file;

    #unlock accounts
    &unlock_file('accounts');

    return "$CONFIG::API{'R_ADDACCT'}:1";
}


# add account type to local DB
sub default_db_add_acct($$) {
    my ($usr, $acct) = @_;
    &$debug("Got $usr, $acct");
    use DBI;
    
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);
    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);
    
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPOSE{'DCF'});
    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    $sth = $dbh->prepare("REPLACE user_account VALUES(?,?)") ||
	return &mysql_error($CONFIG::RESPOSE{'DPF'});
    $sth->execute($acct,$usr) ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    $sth->finish();
    $dbh->disconnect();
    return "$CONFIG::API{'R_ADDACCT'}:1";

}



# remove an account type from a user.
# takes username and account type
sub default_rem_acct ($$) {
    my ($user, $acct) = @_;
    &$debug("Got $user, $acct");
    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $user, $acct\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DREMACCT'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
        &$debug("spawn local $user, $acct\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LREMACCT'};
        return &$handler_sub(@_);
    #
    } else {
        # 
        $target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }
        

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $user, $acct\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'DREMACCT'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);
}

# send remote remove acct call to the appropriate host
sub default_remote_rem_acct ($$$$) { 
    return &RatsClientLib::send_remote_command("REMACCT",@_);
}

# remove account type to local file
sub default_local_rem_acct ($$) {
    my ($user,$acct) = @_;
    &$debug("Got $user\n");
    my $acc_file = $CONFIG::ACCT_FILES{'accounts'};
    my $acc_temp = $CONFIG::ACCT_FILES_TMP{'accounts'};

    #lock the accounts file
    &$debug("locking accounts\n");
    my $result = &lock_file('accounts');
    unless ($result eq $CONFIG::RESPONSE{'SCS'}) {
	&unlock_file('accounts');
	&$log("Failed accounts lock: $result!\n");
	return "$CONFIG::API{'CODEDERR'}:$result";
    }

    #open temp file
    unless (open(TEMP, ">$acc_temp")) {
	&unlock_file('accounts');
	&$log("Failed accounts open: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'ATE'}";
    }

    #copy existing data if any
    if (-f $acc_file) {   
	unless (open(ACCT, $acc_file)) {
	    &$log("Failed accounts open: $!\n");
	    return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'ARE'}";
	}
	# copy existsing data
	while (<ACCT>) { 
	    chomp;
	    my($oldu, $olda) =  split(":");
	    unless ($oldu eq $user && $olda eq $acct) {
		print TEMP "$_\n";
	    }
	}
    }

    # close all file
    close ACCT; close TEMP;

    # move reserve file into place
    tmove ($acc_temp, $acc_file);
    chmod oct $CONFIG::ACCT_FILES_PERMS{'accounts'}, $acc_file;
    chown $CONFIG::ACCT_FILES_OWNER{'accounts'}, $CONFIG::ACCT_FILES_GROUP{'accounts'}, $acc_file;

    #unlock accounts
    &unlock_file('accounts');

    return "$CONFIG::API{'R_REMACCT'}:1";

}

# remove account type to local DB
sub default_db_rem_acct($$) {
    my ($usr, $acct) = @_;
    &$debug("Got $usr, $acct");
    use DBI;
    
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);
    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);
    
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPOSE{'DCF'});
    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    $sth = $dbh->prepare("DELETE FROM user_account WHERE type=? AND uname=?")
	 || return &mysql_error($CONFIG::RESPOSE{'DPF'});
    $sth->execute($acct,$usr) ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->finish();
    $dbh->disconnect();
    return "$CONFIG::API{'R_REMACCT'}:1";

}




# get the account types of a user
# takes username 
sub default_get_acct ($) {
    my ($user) = @_;
    &$debug("Got $user");
    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $user\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DGETACCT'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
        &$debug("spawn local $user\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LGETACCT'};
        return &$handler_sub(@_);
    #
    } else {
        # 
        $target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }
        

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $user\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RGETACCT'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);
}

# send remote get acct call to the appropriate host
sub default_remote_get_acct ($$$) { 
    return &RatsClientLib::send_remote_command("GETACCT",@_);
}

# get account type list from local file
sub default_local_get_acct ($) {
    my ($user) = @_;
    my @list;
    &$debug("Got $user\n");
    my $acc_file = $CONFIG::ACCT_FILES{'accounts'};

    # find entries
    if (-f $acc_file) {   
	unless (open(ACCT, $acc_file)) {
	    &$log("Failed accounts open: $!\n");
	    return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'ARE'}";
	}
	# check for user
	while (<ACCT>) { 
	    chomp;
	    my ($oldu, $olda) =  split(":");
	    if ($oldu eq $user) {
		push @list, $olda;
	    }
	}
    } 

    # close all file
    close ACCT;

    my $temp = join(":", @list);

    return "$CONFIG::API{'R_GETACCT'}:$temp";

}


# get account types from local DB
sub default_db_get_acct($) {
    my ($usr) = @_;
    &$debug("Got $usr");
    use DBI;
    
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth,@list,$temp);
    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);
    
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPOSE{'DCF'});
    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    $sth = $dbh->prepare("SELECT type FROM user_account WHERE uname=?")
	 || return &mysql_error($CONFIG::RESPOSE{'DPF'});
    $sth->execute($usr) ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    while (($temp) = $sth->fetchrow_array()) {
	push @list, $temp;
    }
    $temp = join(":", @list);
    $sth->finish();
    $dbh->disconnect();
    return "$CONFIG::API{'R_GETACCT'}:$temp";

}


# default change gecos call. It figures out
# where the passwd file resides and dispatches the call
sub default_change_gcos ($$$) {
    my ($usr,$gcos,$timestamp) = @_;

 
    &$debug("Got $usr $gcos $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $gcos $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGGCOS'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
        &$debug("spawn local $usr $gcos $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGGCOS'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $gcos $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGGCOS'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change gcos call to appropriate host
sub default_remote_change_gcos (@) {
    return &send_remote_command("CHGGCOS",@_);
}

# changes the gcos of a user by editing passwd file
sub default_local_change_gcos ($$$) {
    my ($target_user, $new_gcos, $timestamp) = @_;
    &$debug("got $target_user, $new_gcos $timestamp\n");
    my($foo);
    $foo = default_change_local_passwd_file($target_user,'gcos',$new_gcos,$CONFIG::API{'R_CHGGCOS'});
    &$debug($foo);
    return $foo;

}


# reserves gcos in local DB by adding username and uid only with
# inactive flag in the password_file table
sub default_db_change_gcos (@) {
    my ($target_user,$new_gcos,$timestamp) = @_;
    my($lock,$timeout,$uid,$local_timestamp);
    use DBI;
    &$debug("got $target_user, $new_gcos $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$target_user\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT uid, lastchange FROM password_file WHERE uname = \"$target_user\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($uid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE password_file set gcos = \"$new_gcos\", lastchange = NULL where uname = \"$target_user\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    # get the timestamp
    my $newt = &get_passlast($dbh, $target_user);
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGGCOS'}:$newt";
}



# default change home call. It figures out
# where the passwd file resides and dispatches the call
sub default_change_home ($$$) {
    my ($usr,$home,$timestamp) = @_;

 
    &$debug("Got $usr $home $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $home $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGHOME'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
        &$debug("spawn local $usr $home $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGHOME'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $home $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGHOME'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change home call to appropriate host
sub default_remote_change_home (@) {
    return &send_remote_command("CHGHOME",@_);
}

# changes the home of a user by editing passwd file
sub default_local_change_home ($$$) {
    my ($target_user, $new_home, $timestamp) = @_;
    &$debug("got $target_user, $new_home $timestamp\n");

    return default_change_local_passwd_file($target_user,'home',$new_home,$CONFIG::API{'R_CHGHOME'});

}


# reserves uid in local DB by adding username and uid only with
# inactive flag in the password_file table
sub default_db_change_home (@) {
    my ($target_user,$new_home,$timestamp) = @_;
    my($lock,$timeout,$uid,$local_timestamp);
    use DBI;
    &$debug("got $target_user, $new_home $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$target_user\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT uid, lastchange FROM password_file WHERE uname = \"$target_user\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($uid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE password_file set home = \"$new_home\", lastchange = NULL where uname = \"$target_user\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    # get the timestamp
    my $newt = &get_passlast($dbh, $target_user);
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGHOME'}:$newt";
}


######## change passpw

# default change passpw call. It figures out
# where the passwd file resides and dispatches the call
sub default_change_passpw ($$$) {
    my ($usr,$passpw,$timestamp) = @_;

 
    &$debug("Got $usr $passpw $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $passpw $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGPASSPW'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
        &$debug("spawn local $usr $passpw $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGPASSPW'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $passpw $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGPASSPW'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change password  call to appropriate host
sub default_remote_change_passpw (@) {
    return &send_remote_command("CHGPASSPW",@_);
}

# changes the password of a user by editing passwd file
sub default_local_change_passpw ($$$) {
    my ($target_user, $new_passpw, $timestamp) = @_;
    &$debug("got $target_user, $new_passpw $timestamp\n");

    return default_change_local_passwd_file($target_user,'pass',$new_passpw,$CONFIG::API{'R_CHGPASSPW'});

}


# change the password in the password file for a given uname. 
# database version
sub default_db_change_passpw (@) {
    my ($target_user,$new_passpw,$timestamp) = @_;
    my($lock,$timeout,$uid,$local_timestamp);
    use DBI;
    &$debug("got $target_user, $new_passpw $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$target_user\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT uid, lastchange FROM password_file WHERE uname = \"$target_user\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($uid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE password_file set pass = \"$new_passpw\", lastchange = NULL where uname = \"$target_user\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    # get the timestamp
    my $newt = &get_passlast($dbh, $target_user);
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGPASSPW'}:$newt";
}

######## change uid


# default change uid call. It figures out
# where the passwd file resides and dispatches the call
sub default_change_uid ($$$) {
    my ($usr,$uid,$timestamp) = @_;

 
    &$debug("Got $usr $uid $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $uid $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGUID'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
        &$debug("spawn local $usr $uid $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGUID'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $uid $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGUID'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change uid call to appropriate host
sub default_remote_change_uid (@) {
    return &send_remote_command("CHGUID",@_);
}

# changes the uid of a user by editing passwd file
sub default_local_change_uid ($$$) {
    my ($target_user, $new_uid, $timestamp) = @_;
    &$debug("got $target_user, $new_uid $timestamp\n");
    default_change_local_reserve_file($target_user,'uid',$new_uid,$CONFIG::API{'R_CHGUID'});
    return default_change_local_passwd_file($target_user,'uid',$new_uid,$CONFIG::API{'R_CHGUID'});
    
}


# change the uid in the password file for a given uname. 
# database version
sub default_db_change_uid (@) {
    my ($target_user,$new_uid,$timestamp) = @_;
    my($lock,$timeout,$uid,$local_timestamp);
    use DBI;
    &$debug("got $target_user, $new_uid $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$target_user\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT uid, lastchange FROM password_file WHERE uname = \"$target_user\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($uid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    $new_uid++; #increment by one for the mysql autoincrement thing
    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE password_file set uid = \"$new_uid\", lastchange = NULL where uname = \"$target_user\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    # get the timestamp
    my $newt = &get_passlast($dbh, $target_user); 
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGUID'}:$newt";
}

######## change gidpw

# default change gidpw call. It figures out
# where the passwd file resides and dispatches the call
sub default_change_gidpw ($$$) {
    my ($usr,$gidpw,$timestamp) = @_;

 
    &$debug("Got $usr $gidpw $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $gidpw $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGGIDPW'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
        &$debug("spawn local $usr $gidpw $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGGIDPW'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $gidpw $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGGIDPW'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change gidpw call to appropriate host
sub default_remote_change_gidpw (@) {
    return &send_remote_command("CHGGIDPW",@_);
}

# changes the gidpw of a user by editing passwd file
sub default_local_change_gidpw ($$$) {
    my ($target_user, $new_gidpw, $timestamp) = @_;
    &$debug("got $target_user, $new_gidpw $timestamp\n");
    return default_change_local_passwd_file($target_user,'gid',$new_gidpw,$CONFIG::API{'R_CHGGIDPW'});
}


# change the gid in the password file for a given uname. 
# database version
sub default_db_change_gidpw (@) {
    my ($target_user,$new_gidpw,$timestamp) = @_;
    my($lock,$timeout,$gidpw,$local_timestamp,$uid);
    use DBI;
    &$debug("got $target_user, $new_gidpw $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$target_user\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT uid, lastchange FROM password_file WHERE uname = \"$target_user\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($uid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    # user exitsts, timestamp is good, so update the row
    $new_gidpw++;
    $dbh->do("UPDATE password_file set gid = \"$new_gidpw\", lastchange = NULL where uname = \"$target_user\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    # get the timestamp
    my $newt = &get_passlast($dbh, $target_user);
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGGIDPW'}:$newt";
}

#### change uname

# default change unamepw call. It figures out
# where the passwd file resides and dispatches the call
sub default_change_uname ($$$) {
    my ($usr,$uname,$timestamp) = @_;

 
    &$debug("Got $usr $uname $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $uname $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGUNAME'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
        &$debug("spawn local $usr $uname $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGUNAME'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $uname $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGUNAME'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change uname call to appropriate host
sub default_remote_change_uname (@) {
    return &send_remote_command("CHGUNAME",@_);
}

# changes the uname of a user by editing passwd file
sub default_local_change_uname ($$$) {
    my ($target_user, $new_uname, $timestamp) = @_;
    &$debug("got $target_user, $new_uname $timestamp\n");
    default_change_local_reserve_file($target_user,'uname',$new_uname,$CONFIG::API{'R_CHGUNAME'});
    default_change_local_shadow_file($target_user,'uname',$new_uname,$CONFIG::API{'R_CHGUNAME'});
    return default_change_local_passwd_file($target_user,'uname',$new_uname,$CONFIG::API{'R_CHGUNAME'});
     
}


# change the uname in the password file for a given uname. 
# database version
sub default_db_change_uname (@) {
    my ($target_user,$new_uname,$timestamp) = @_;
    my($lock,$timeout,$uname,$local_timestamp,$uid);
    use DBI;
    &$debug("got $target_user, $new_uname $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$target_user\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT uid, lastchange FROM password_file WHERE uname = \"$target_user\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($uid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE password_file set uname = \"$new_uname\", lastchange = NULL where uname = \"$target_user\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    # get the timestamp
    my $newt = &get_passlast($dbh, $new_uname);
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGUNAME'}:$newt";
}


#### group change gname

# default change gnameu call. It figures out
# where the group file resides and dispatches the call
sub default_change_gname ($$$) {
    my ($group,$gname,$timestamp) = @_;

 
    &$debug("Got $group $gname $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $group $gname $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGGNAME'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'group'} eq $host) {
        &$debug("spawn local $group $gname $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGGNAME'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'group'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $group $gname $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGGNAME'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change gname call to appropriate host
sub default_remote_change_gname (@) {
    return &send_remote_command("CHGGNAME",@_);
}

# changes the gname of a user by editing group file
sub default_local_change_gname ($$$) {
    my ($target_group, $new_gname, $timestamp) = @_;
    &$debug("got $target_group, $new_gname $timestamp\n");
    default_change_local_reserveg_file($target_group,'gname',$new_gname,$CONFIG::API{'R_CHGGNAME'});
    return default_change_local_group_file($target_group,'gname',$new_gname,$CONFIG::API{'R_CHGGNAME'});
}


# change the gname in the group file for a given gname. 
# database version
sub default_db_change_gname (@) {
    my ($target_group,$new_gname,$timestamp) = @_;
    my($lock,$timeout,$gid,$local_timestamp);
    use DBI;
    &$debug("got $target_group, $new_gname, $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"group_file-$target_group\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT gid, lastchange FROM group_file WHERE gname = \"$target_group\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$gid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($gid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE group_file set gname = \"$new_gname\", lastchange = NULL where gname = \"$target_group\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    $dbh->disconnect();
    # get the timestamp
    my $newt = &get_grplast($dbh, $new_gname);
    return "$CONFIG::API{'R_CHGGNAME'}:$newt";
}


#### group change password passgr

# default change passgr call. It figures out
# where the group file resides and dispatches the call
sub default_change_passgr ($$$) {
    my ($group,$passgr,$timestamp) = @_;

 
    &$debug("Got $group $passgr $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $group $passgr $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGPASSGR'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'group'} eq $host) {
        &$debug("spawn local $group $passgr $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGPASSGR'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'group'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $group $passgr $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGPASSGR'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change passgr call to appropriate host
sub default_remote_change_passgr (@) {
    return &send_remote_command("CHGPASSGR",@_);
}

# changes the passgr of a user by editing group file
sub default_local_change_passgr ($$$) {
    my ($target_group, $new_passgr, $timestamp) = @_;
    &$debug("got $target_group, $new_passgr $timestamp\n");

    return default_change_local_group_file($target_group,'pass',$new_passgr,$CONFIG::API{'R_CHGPASSGR'});
}


# change the passgr in the group file for a given gname. 
# database version
sub default_db_change_passgr (@) {
    my ($target_group,$new_passgr,$timestamp) = @_;
    my($lock,$timeout,$gid,$local_timestamp);
    use DBI;
    &$debug("got $target_group, $new_passgr, $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"group_file-$target_group\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT gid, lastchange FROM group_file WHERE gname = \"$target_group\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$gid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($gid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE group_file set pass = \"$new_passgr\", lastchange = NULL where gname = \"$target_group\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    # get the timestamp
    my $newt = &get_grplast($dbh, $target_group);
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGPASSGR'}:$newt";
}

#### group change gid gidgr

# default change gidgr call. It figures out
# where the group file resides and dispatches the call
sub default_change_gidgr ($$$) {
    my ($group,$gidgr,$timestamp) = @_;

 
    &$debug("Got $group $gidgr $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $group $gidgr $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGGIDGR'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'group'} eq $host) {
        &$debug("spawn local $group $gidgr $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGGIDGR'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'group'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $group $gidgr $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGGIDGR'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change gidgr call to appropriate host
sub default_remote_change_gidgr (@) {
    return &send_remote_command("CHGGIDGR",@_);
}

# changes the gidgr of a user by editing group file
sub default_local_change_gidgr ($$$) {
    my ($target_group, $new_gidgr, $timestamp) = @_;
    &$debug("got $target_group, $new_gidgr $timestamp\n");
    default_change_local_reserveg_file($target_group,'gid',$new_gidgr,$CONFIG::API{'R_CHGGIDGR'});
    return default_change_local_group_file($target_group,'gid',$new_gidgr,$CONFIG::API{'R_CHGGIDGR'});

}


# change the gidgr in the group file for a given gname. 
# database version
sub default_db_change_gidgr (@) {
    my ($target_group,$new_gidgr,$timestamp) = @_;
    my($lock,$timeout,$gid,$local_timestamp);
    use DBI;
    &$debug("got $target_group, $new_gidgr, $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"group_file-$target_group\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT gid, lastchange FROM group_file WHERE gname = \"$target_group\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$gid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($gid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }
    $new_gidgr++;
    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE group_file set gid = \"$new_gidgr\", lastchange = NULL where gname = \"$target_group\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    # get the timestamp
    my $newt = &get_grplast($dbh, $target_group);
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGGIDGR'}:$newt";
}

#### shadow change last

# default change slast call. It figures out
# where the shadow file resides and dispatches the call
sub default_change_slast ($$$) {
    my ($usr,$slast,$timestamp) = @_;

 
    &$debug("Got $usr $slast $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $slast $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGSLAST'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'shadow'} eq $host) {
        &$debug("spawn local $usr $slast $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGSLAST'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'shadow'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $slast $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGSLAST'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change slast call to appropriate host
sub default_remote_change_slast (@) {
    return &send_remote_command("CHGSLAST",@_);
}

# changes the slast of a user by editing shadow file
sub default_local_change_slast ($$$) {
    my ($target_user, $new_slast, $timestamp) = @_;
    &$debug("got $target_user, $new_slast $timestamp\n");

    return default_change_local_shadow_file($target_user,'lastchg',$new_slast,$CONFIG::API{'R_CHGSLAST'});

}


#change slast call executed on DB. changes the value of last in the 
#password_file table (is shadow file data)
sub default_db_change_slast (@) {
    my ($target_user,$new_slast,$timestamp) = @_;
    my($lock,$timeout,$uid,$local_timestamp);
    use DBI;
    &$debug("got $target_user, $new_slast $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$target_user\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT uid, lastchange FROM password_file WHERE uname = \"$target_user\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($uid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE password_file set last = \"$new_slast\", lastchange = NULL where uname = \"$target_user\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    # get the timestamp
    my $newt = &get_passlast($dbh, $target_user);
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGSLAST'}:$newt";
}




#### shadow change min

# default change smin call. It figures out
# where the shadow file resides and dispatches the call
sub default_change_smin ($$$) {
    my ($usr,$smin,$timestamp) = @_;

 
    &$debug("Got $usr $smin $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $smin $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGSMIN'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'shadow'} eq $host) {
        &$debug("spawn local $usr $smin $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGSMIN'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'shadow'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $smin $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGSMIN'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change smin call to appropriate host
sub default_remote_change_smin (@) {
    return &send_remote_command("CHGSMIN",@_);
}

# changes the smin of a user by editing shadow file
sub default_local_change_smin ($$$) {
    my ($target_user, $new_smin, $timestamp) = @_;
    &$debug("got $target_user, $new_smin $timestamp\n");

    return default_change_local_shadow_file($target_user,'mindays',$new_smin,$CONFIG::API{'R_CHGSMIN'});

}


#change smin call executed on DB. changes the value of last in the 
#password_file table (is shadow file data)
sub default_db_change_smin (@) {
    my ($target_user,$new_smin,$timestamp) = @_;
    my($lock,$timeout,$uid,$local_timestamp);
    use DBI;
    &$debug("got $target_user, $new_smin $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$target_user\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT uid, lastchange FROM password_file WHERE uname = \"$target_user\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($uid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE password_file set min = \"$new_smin\", lastchange = NULL where uname = \"$target_user\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    # get the timestamp
    my $newt = &get_passlast($dbh, $target_user);
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGSMIN'}:$newt";
}


#### shadow change max

# default change smax call. It figures out
# where the shadow file resides and dispatches the call
sub default_change_smax ($$$) {
    my ($usr,$smax,$timestamp) = @_;

 
    &$debug("Got $usr $smax $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $smax $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGSMAX'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'shadow'} eq $host) {
        &$debug("spawn local $usr $smax $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGSMAX'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'shadow'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $smax $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGSMAX'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change smax call to appropriate host
sub default_remote_change_smax (@) {
    return &send_remote_command("CHGSMAX",@_);
}

# changes the smax of a user by editing passwd file
sub default_local_change_smax ($$$) {
    my ($target_user, $new_smax, $timestamp) = @_;
    &$debug("got $target_user, $new_smax $timestamp\n");

    return default_change_local_shadow_file($target_user,'maxdays',$new_smax,$CONFIG::API{'R_CHGSMAX'});
 
}


#change smax call executed on DB. changes the value of last in the 
#password_file table (is shadow file data)
sub default_db_change_smax (@) {
    my ($target_user,$new_smax,$timestamp) = @_;
    my($lock,$timeout,$uid,$local_timestamp);
    use DBI;
    &$debug("got $target_user, $new_smax $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$target_user\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT uid, lastchange FROM password_file WHERE uname = \"$target_user\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($uid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE password_file set max = \"$new_smax\", lastchange = NULL where uname = \"$target_user\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    # get the timestamp
    my $newt = &get_passlast($dbh, $target_user);
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGSMAX'}:$newt";
}


#### shadow change warn

# default change swarn call. It figures out
# where the shadow file resides and dispatches the call
sub default_change_swarn ($$$) {
    my ($usr,$swarn,$timestamp) = @_;

 
    &$debug("Got $usr $swarn $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $swarn $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGSWARN'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'shadow'} eq $host) {
        &$debug("spawn local $usr $swarn $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGSWARN'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'shadow'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $swarn $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGSWARN'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change swarn call to appropriate host
sub default_remote_change_swarn (@) {
    return &send_remote_command("CHGSWARN",@_);
}

# changes the swarn of a user by editing shadow file
sub default_local_change_swarn ($$$) {
    my ($target_user, $new_swarn, $timestamp) = @_;
    &$debug("got $target_user, $new_swarn $timestamp\n");

    return default_change_local_shadow_file($target_user,'warn',$new_swarn,$CONFIG::API{'R_CHGSWARN'});
      
}


#change swarn call executed on DB. changes the value of last in the 
#password_file table (is shadow file data)
sub default_db_change_swarn (@) {
    my ($target_user,$new_swarn,$timestamp) = @_;
    my($lock,$timeout,$uid,$local_timestamp);
    use DBI;
    &$debug("got $target_user, $new_swarn $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$target_user\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT uid, lastchange FROM password_file WHERE uname = \"$target_user\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($uid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE password_file set warn = \"$new_swarn\", lastchange = NULL where uname = \"$target_user\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    # get the timestamp
    my $newt = &get_passlast($dbh, $target_user);
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGSWARN'}:$newt";
}


#### shadow change inactive

# default change sinaccall. It figures out
# where the shadow file resides and dispatches the call
sub default_change_sinac ($$$) {
    my ($usr,$sinac,$timestamp) = @_;

 
    &$debug("Got $usr $sinac $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $sinac $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGSINAC'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'shadow'} eq $host) {
        &$debug("spawn local $usr $sinac $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGSINAC'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'shadow'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $sinac $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGSINAC'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change sinac call to appropriate host
sub default_remote_change_sinac (@) {
    return &send_remote_command("CHGSINAC",@_);
}

# changes the sinac of a user by editing shadow file
sub default_local_change_sinac ($$$) {
    my ($target_user, $new_sinac, $timestamp) = @_;
    &$debug("got $target_user, $new_sinac $timestamp\n");

    return default_change_local_shadow_file($target_user,'inactive',$new_sinac,$CONFIG::API{'R_CHGSINAC'});
        
}


#change sinac call executed on DB. changes the value of last in the 
#password_file table (is shadow file data)
sub default_db_change_sinac (@) {
    my ($target_user,$new_sinac,$timestamp) = @_;
    my($lock,$timeout,$uid,$local_timestamp);
    use DBI;
    &$debug("got $target_user, $new_sinac $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$target_user\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT uid, lastchange FROM password_file WHERE uname = \"$target_user\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($uid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE password_file set inactive = \"$new_sinac\", lastchange = NULL where uname = \"$target_user\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    # get the timestamp
    my $newt = &get_passlast($dbh, $target_user);
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGSINAC'}:$newt";
}


#### shadow change expire

# default change sexp call. It figures out
# where the shadow file resides and dispatches the call
sub default_change_sexp ($$$) {
    my ($usr,$sexp,$timestamp) = @_;

 
    &$debug("Got $usr $sexp $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $sexp $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGSEXP'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'shadow'} eq $host) {
        &$debug("spawn local $usr $sexp $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGSEXP'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'shadow'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $sexp $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGSEXP'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change sexp call to appropriate host
sub default_remote_change_sexp (@) {
    return &send_remote_command("CHGSEXP",@_);
}

# changes the sexp of a user by editing shadow file
sub default_local_change_sexp ($$$) {
    my ($target_user, $new_sexp, $timestamp) = @_;
    &$debug("got $target_user, $new_sexp $timestamp\n");

    return default_change_local_shadow_file($target_user,'expire',$new_sexp,$CONFIG::API{'R_CHGSEXP'});
}


#change sexp call executed on DB. changes the value of last in the 
#password_file table (is shadow file data)
sub default_db_change_sexp (@) {
    my ($target_user,$new_sexp,$timestamp) = @_;
    my($lock,$timeout,$uid,$local_timestamp);
    use DBI;
    &$debug("got $target_user, $new_sexp $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$target_user\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT uid, lastchange FROM password_file WHERE uname = \"$target_user\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($uid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE password_file set expire = \"$new_sexp\", lastchange = NULL where uname = \"$target_user\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    # get the timestamp
    my $newt = &get_passlast($dbh, $target_user);
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGSEXP'}:$newt";
}

#### shadow change flag

# default change sflag call. It figures out
# where the shadow file resides and dispatches the call
sub default_change_sflag ($$$) {
    my ($usr,$sflag,$timestamp) = @_;

 
    &$debug("Got $usr $sflag $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $sflag $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGSFLAG'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'shadow'} eq $host) {
        &$debug("spawn local $usr $sflag $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGSFLAG'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'shadow'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $sflag $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGSFLAG'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change sflag call to appropriate host
sub default_remote_change_sflag (@) {
    return &send_remote_command("CHGSFLAG",@_);
}

# changes the sflag of a user by editing shadow file
sub default_local_change_sflag ($$$) {
    my ($target_user, $new_sflag, $timestamp) = @_;
    &$debug("got $target_user, $new_sflag $timestamp\n");

        return default_change_local_shadow_file($target_user,'flag',$new_sflag,$CONFIG::API{'R_CHGSFLAG'});
}


#change sflag call executed on DB. changes the value of last in the 
#password_file table (is shadow file data)
sub default_db_change_sflag (@) {
    my ($target_user,$new_sflag,$timestamp) = @_;
    my($lock,$timeout,$uid,$local_timestamp);
    use DBI;
    &$debug("got $target_user, $new_sflag $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$target_user\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT uid, lastchange FROM password_file WHERE uname = \"$target_user\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($uid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE password_file set flag = \"$new_sflag\", lastchange = NULL where uname = \"$target_user\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    # get the timestamp
    my $newt = &get_passlast($dbh, $target_user);
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGSFLAG'}:$newt";
}


#### shadow password

# default change spw call. It figures out
# where the passwd file resides and dispatches the call
sub default_change_spw ($$$) {
    my ($usr,$spw,$timestamp) = @_;

 
    &$debug("Got $usr $spw $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $spw $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGSPW'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'shadow'} eq $host) {
        &$debug("spawn local $usr $spw $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGSPW'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'shadow'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $spw $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGSPW'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change password  call to appropriate host
sub default_remote_change_spw (@) {
    return &send_remote_command("CHGSPW",@_);
}

# changes the password of a user by editing shadow file
sub default_local_change_spw ($$$) {
    my ($target_user, $new_spw, $timestamp) = @_;
    &$debug("got $target_user, $new_spw $timestamp\n");

    return default_change_local_shadow_file($target_user,'pass',$new_spw,$CONFIG::API{'R_CHGSPW'});

}


# change the password in the password file for a given uname. 
# database version
sub default_db_change_spw (@) {
    my ($target_user,$new_spw,$timestamp) = @_;
    my($lock,$timeout,$uid,$local_timestamp);
    use DBI;
    &$debug("got $target_user, $new_spw $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$target_user\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT uid, lastchange FROM password_file WHERE uname = \"$target_user\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($uid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE password_file set pass = \"$new_spw\", lastchange = NULL where uname = \"$target_user\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    # get the timestamp
    my $newt = &get_passlast($dbh, $target_user);
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGSPW'}:$newt";
}

#### shadow uname change

# default change suname call. It figures out
# where the passwd file resides and dispatches the call
sub default_change_suname ($$$) {
    my ($usr,$suname,$timestamp) = @_;

 
    &$debug("Got $usr $suname $timestamp\n");    
    my $host = &get_hostname();
    
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB $usr $suname $timestamp\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHGSUNAME'};
            return &$handler_sub(@_);
        # if remote then sety $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'shadow'} eq $host) {
        &$debug("spawn local $usr $suname $timestamp\n");
        my $handler_sub = $CONFIG::COMMAND_HOOKS{'LCHGSUNAME'};
        return &$handler_sub(@_);
    #
    } else {
        #
        $target_host = $CONFIG::ACCT_FILES_MASTER{'shadow'};
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote $usr $suname $timestamp\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHGSUNAME'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);

}


# Dispatches the change suname call to appropriate host
sub default_remote_change_suname (@) {
    return &send_remote_command("CHGSUNAME",@_);
}

# changes the suname of a user by editing shadow file
sub default_local_change_suname ($$$) {
    my ($target_user, $new_suname, $timestamp) = @_;
    &$debug("got $target_user, $new_suname $timestamp\n");
    default_change_local_reserve_file($target_user,'uname',$new_suname,$CONFIG::API{'R_CHGSUNAME'});
    default_change_local_passwd_file($target_user,'uname',$new_suname,$CONFIG::API{'R_CHGSUNAME'});
    return default_change_local_shadow_file($target_user,'uname',$new_suname,$CONFIG::API{'R_CHGSUNAME'});
    
}


# change the suname in the password file for a given uname. 
# database version
sub default_db_change_suname (@) {
    my ($target_user,$new_suname,$timestamp) = @_;
    my($lock,$timeout,$suname,$local_timestamp,$uid);
    use DBI;
    &$debug("got $target_user, $new_suname $timestamp\n");

    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }
    # lock entry while we much around. it goes away at disconnect
    $timeout = $CONFIG::SLEEP_LEN * $CONFIG::LOCK_TRIES;
    $sth = $dbh->prepare("SELECT GET_LOCK(\"password_file-$target_user\",\"$timeout\")") ||
        return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});
    ($lock) = $sth->fetchrow_array();
    if ($lock == 0) {
        $sth->finish();
        $dbh->disconnect();
        return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DLF'}";
    } elsif ($lock eq "") {
        $sth->finish();
        $dbh->disconnect();
        return &mysql_error("$CONFIG::RESPONSE{'DEF'}:$DBI::err");
    }
    

    # check if user already exists
    $sth = $dbh->prepare("SELECT uid, lastchange FROM password_file WHERE uname = \"$target_user\"")
	||return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() || return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns (\$uid,\$local_timestamp);
    $sth->fetch();
    $sth->finish();
    unless($uid) {
	$dbh->disconnect();
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
    unless($local_timestamp == $timestamp || $timestamp eq 0){
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'TDM'}";
    }

    # user exitsts, timestamp is good, so update the row
    $dbh->do("UPDATE password_file set uname = \"$new_suname\", lastchange = NULL where uname = \"$target_user\"") ||
        return &mysql_error($CONFIG::RESPONSE{'DEF'});

    #clean up
    # get the timestamp
    my $newt = &get_passlast($dbh, $new_suname);
    $dbh->disconnect();
    return "$CONFIG::API{'R_CHGSUNAME'}:$newt";
}






#stub function for changing fields in passwd file
sub default_change_local_passwd_file ($$$$) {
    &$debug("got @_");
    default_change_local_file('passwd', \%passwd_name2num, @_);
}

#stub function for changing fields in reserve file
sub default_change_local_reserve_file ($$$$) {
    &$debug("got @_");
    default_change_local_file('reserve', \%reserve_name2num, @_);

}

#stub function for changing fields in reserveg file
sub default_change_local_reserveg_file ($$$$) {
    &$debug("got @_");
    default_change_local_file('reserveg', \%reserveg_name2num, @_);
}

#stub function for changing fields in shadow file
sub default_change_local_shadow_file ($$$$) {
    &$debug("got @_");
    default_change_local_file('shadow', \%shadow_name2num, @_);
}

#stub function for changing fields in group file
sub default_change_local_group_file ($$$$) {
    &$debug("got @_");
    default_change_local_file('group', \%group_name2num, @_);
}


# changes the field of a user by editing a local file
# this supports most standard ':' delimited files
sub default_change_local_file ($\%$$$) {
  my ($filename, $fieldhash, $target_user, $field_name, $new_value, $ret) = @_;
  my ($response, $found, $TEMP);
  &$debug("for $filename got $target_user, $field_name, $new_value\n");

  unless (exists $fieldhash->{$field_name}) {
    &$log("illegal $filename field name specified: $response!\n");
    return "$CONFIG::API{'CODEDERR'}:$field_name is bogus $filename field";
  }

  eval {			# catch exception and cleanup
    #lock local file
    $response = lock_file("$filename");
    unless ($response eq $CONFIG::RESPONSE{'SCS'}) {
      &$log("Failed $filename lock: $response!\n");
      $@ = $response;
      return;
    }

    # open local file
    my $LOCAL = $CONFIG::ACCT_FILES{"$filename"};
    unless (open(LOCAL, "<$LOCAL")) {
      unlock_file("$filename");
      &$log("Failed $filename open: $!\n");
      $@ = "$CONFIG::RESPONSE{'PRE'}";
      return;
    }

    # open temp file
    $TEMP = $CONFIG::ACCT_FILES_TMP{"$filename"};
    unless (open(TEMP, ">$TEMP")) {
      unlock_file("$filename");
      &$log("Failed $filename temp open: $!\n");
      $@ = "$CONFIG::RESPONSE{'PTE'}";
      return;
    }

    # do the change and keep track if you found the user
    $found = 0;
    {
      my $field;
      my @fields;
      while (<LOCAL>) {
	chomp;
	# append the POST so null fields get included
	@fields = split(':', $_ . ':POST');
	# should check if entry parses correctly, but don't :-(!!!
	if ($fields[0] ne $target_user) {
	  print TEMP "$_\n";
	  next;
	}
	pop @fields;		# discard POST field
	$field = $fieldhash->{$field_name};
	$fields[$field] = $new_value;
	print TEMP join(':', @fields),"\n";
	$found = 1;
	# allow multuple entries for this username (even though bogus)
      }
    }

    close LOCAL;		# error test????!!!!
    close TEMP;

    # and we move it in the right place
    tmove($TEMP, $LOCAL);
    chmod oct $CONFIG::ACCT_FILES_PERMS{"$filename"}, $LOCAL;

    # and unlock local file
    unlock_file("$filename");
  };
  if ($@) {
    # something wrong (or exception) so clean up lock and temp file
    my $msg = $@;		# save current error message
    close LOCAL;
    close TEMP;
    if (-f $TEMP) {unlink "$TEMP"}
    eval {unlock_file("$filename")}; # ignore any errors on an unlock
    return "$CONFIG::API{'CODEDERR'}:$msg"; # propogate error message
  }

  if ($found) {
    &$debug("$target_user $field_name chgd $new_value\n");
    return "$ret:1";
  } else {
    &$log("$target_user not found\n");
    return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
  }
}




# the default generate gid function
# WARNING: this function assumes that you already have the group file
# WARNING: locked. Otherwise multiple proccess could get the same gid 
# WARNING: before the first one finished updating the file
#
# it first checks if GID syncronization if being done. If so it contacts 
# the appropriate remote host. If this is running on a GID sync master 
# then it generates a new uid.

sub default_generate_gid  ($) {
    my ($result, $code, @data, $handler_sub, $remote, $port);
    my ($res_file, $gid, $grp);
    my $group = $_[0];
    my $host = &get_hostname();

    # get remote, first try to look it up, then reserve it
    if ($CONFIG::GID_SYNC && $CONFIG::GID_HOST ne $host) {
	&$debug("trying to obtained remote gid for $group\n");
	#getuid from remote host
	$result = &$do('GETGID', $group, $CONFIG::GID_HOST, $CONFIG::CLIENT_PORT);
	($code, @data) = split(":", $result);
	if ($code == $CONFIG::API{'R_GETGID'}) {      # try to get it for existing user
	    return $data[0];
	} elsif ($code == $CONFIG::API{'NOTFOUND'}) { # reserve uid
	    &$debug("trying to reserve remote gid for $group\n");
	    $result = &$do('RESERVEG', $group, $CONFIG::GID_HOST, $CONFIG::CLIENT_PORT);
	    ($code, @data) = split(":", $result);
	    if ($code == $CONFIG::API{'R_RESERVEG'}) { 
		return $data[0];
	    } else {
		&$log("failed to reserve gid for $group: @data\n");
		return $CONFIG::BOGUS_GID;
	    }
	} else {
	    &$log("failed to get gid for $group: @data\n");
	    return $CONFIG::BOGUS_GID;
	}

    } else {    # get local

	&$debug("trying to obtained local gid for $group\n");
	# first we check if was already reserved
	$res_file = $CONFIG::ACCT_FILES{'reserveg'};
	if ( -f $res_file) {
	    unless (open(RESERVE, $res_file)) {
		&$log("Failed to open reserve file: $!\n");
		return $CONFIG::BOGUS_GID;
	    }
	    while (<RESERVE>) {
		chomp;
		($grp, $gid) = split(":", $_);
		if ($group eq $grp) {
		    return $gid;
		}
	    }
	    close RESERVE;
	}
	# otherwise we generate one
	$handler_sub = $CONFIG::COMMAND_HOOKS{'GENNGID'};
	return &$handler_sub($group);
    }
	
}

# the default generate gid function return the next highes avaible gid
# based on the current group file. 
#
# WARNING: this function assumes that you already have the group file
# WARNING: locked. Otherwise multiple proccess could get the same gid 
# WARNING: before the first one finished updating the file
# 
# Note that it also checks the reserved file to make sure that it catches 
# all the pre allocated GID's
#
# it returns the uid or BOGUS_GID for failure
sub default_generate_next_gid  ($) {
    my $arg = $_[0];
    my ($max_gid, $gid, $entry, $grp_file, @group);
    my ($grp, $pass, @rest, $res_file, $res_temp);
    my (@reserve, %gid_space);

    # the max uid (which is what we are going to add 1 to and return
    # can't be less the MIN_GID - 1 
    $max_gid = $CONFIG::MIN_GID - 1;

    # read the group file
    $grp_file = $CONFIG::ACCT_FILES{'group'};
    unless (open(GRP, $grp_file)) {
	&$log("Failed to open group file: $!\n");
	return $CONFIG::BOGUS_GID;
    }
    @group =  <GRP>;
    close GRP;

    # read the reserve file
    if ($CONFIG::ACCT_FILES{'reserveg'} && -f $CONFIG::ACCT_FILES{'reserveg'}){
	$res_file = $CONFIG::ACCT_FILES{'reserveg'};
	unless (open(RESERVE, $res_file)) {
	    &$log("Failed to open reserve file: $!\n");
	    return $CONFIG::BOGUS_GID;
	}
	@reserve = <RESERVE>;
	close RESERVE;
    }


    if ($CONFIG::USE_NEXT_GID) { # get the next gid
	&$debug("looking for first free\n");
	# find the groups gid's
	foreach $entry (@group) {
	    ($grp, $pass, $gid, @rest) = split(":", $entry);
	    $gid_space{$gid} = 1;
	}
	foreach $entry (@reserve) {
	    ($grp, $gid) = split(":", $entry);
	    $gid_space{$gid} = 1;
	}
	for ($max_gid = $CONFIG::MIN_GID;;$max_gid++) {
	       last unless $gid_space{$max_gid};
	}

    } else {  	# find the max gid
	&$debug("looking for largest free\n");
	foreach $entry (@group) {
	    ($grp, $pass, $gid, @rest) = split(":", $entry);
	    if ($gid > $max_gid) {
		$max_gid = $gid;
	    }
	}
	
	#check for max in reserved file
	# find the max gid
	foreach $entry (@reserve) {
	    chomp $entry;
	    ($grp, $gid) = split(":", $entry);
	    if ($gid > $max_gid) {
		$max_gid = $gid;
	    }
	}
	# generate the next uid
	$max_gid++;
    }
    &$debug("next gid is $max_gid\n");
    # update reserve file
    $res_temp = $CONFIG::ACCT_FILES_TMP{'reserveg'};
    unless (open(TEMP, ">$res_temp")) {
	&$log("Failed reserve open: $!\n");
	return $CONFIG::BOGUS_GID;
    }
    
    $res_file = $CONFIG::ACCT_FILES{'reserveg'};
    if (-f $res_file) {   
	unless (open(RESERVE, $res_file)) {
	    &$log("Failed reserve open: $!\n");
	    return $CONFIG::BOGUS_GID;
	}
	# copy existsing data
	while (<RESERVE>) { print TEMP;}
    }

    # put in new entry
    print TEMP "$arg:$max_gid\n";

    # close all file
    close RESERVE; close TEMP;

    # move reserve file into place
    tmove ($res_temp, $res_file);
    chmod oct $CONFIG::ACCT_FILES_PERMS{'reserveg'}, $res_file;
    chown $CONFIG::ACCT_FILES_OWNER{'reserveg'}, $CONFIG::ACCT_FILES_GROUP{'reserveg'}, $res_file;

    # return the next highest uid
    return $max_gid;
}

# get the time stamp of the last update on named field
# first arg must be a open dbh
sub get_passlast ($$) {
    my ($dbh, $user) = @_;
    my $rval;
    my $sth1=$dbh->prepare("SELECT lastchange FROM password_file WHERE uname = ?") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth1->execute($user) || 
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    while (($rval) = $sth1->fetchrow_array()) {
        return $rval;
    }
    return 0;
}

# get the time stamp of the last update on named field
# first arg must be a open dbh
sub get_grplast ($$) {
    my ($dbh, $grp) = @_;
    my $rval;
    my $sth1=$dbh->prepare("SELECT lastchange FROM group_file WHERE gname = ?") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth1->execute($grp) || 
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    while (($rval) = $sth1->fetchrow_array()) {
        return $rval;
    }
    return 0;
}


sub db_global_unlock($)
{
use DBI;
my($dbh) = $_[0];
my ($sth, $gid, $count,$lock,$timeout);

$sth = $dbh->prepare("delete from rats_locks where"
		     ." t_name = ?") ||
			return &mysql_error("$CONFIG::RESPONSE{'DPF'}");
$sth->execute($CONFIG::USE_CLIENT_DB)||
    &mysql_error($CONFIG::RESPONSE{'DEF'});
$sth->finish();
return;
}


sub db_global_lock($)
{
    use DBI;
    my($dbh) = $_[0];
    my ($sth, $gid, $count,$lock,$timeout,$global_lock);
    
    
#check global lock
    
    $global_lock=0;
    $sth = $dbh->prepare("select count(*) from rats_locks where"
			 ." t_name = ?") || 
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute($CONFIG::USE_CLIENT_DB)||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    $sth->bind_columns(\$global_lock);
    $sth->fetch();
    $sth->finish();

#if locked, puke
    if($global_lock)
    {
#nasty use of mysql_error... but it is an artificial
#rats specific db error...
	return &mysql_error("Global RATS advisory lock already in place\n");
	
    }

#no global lock, so set your own..
    
    $sth = $dbh->prepare("INSERT rats_locks VALUES(?,'read',NULL)") ||
	return &mysql_error("$CONFIG::RESPONSE{'DPF'}");
    $sth->execute($CONFIG::USE_CLIENT_DB)||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});
    
    $sth->finish();
    $dbh->disconnect();
    return;
}



# check if the db is locked in an API context
sub default_check_db_lock() {
    &$debug("Got query.\n");    
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");
    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
        # if local then do it
        if ($CONFIG::CLIENT_DB_HOST eq $host) {
            &$debug("spawn DB query\n");
            my $handler_sub = $CONFIG::COMMAND_HOOKS{'DCHKDBLOCK'};
            return &$handler_sub();
        # if remote then set $target_host
        } else {
            $target_host = $CONFIG::CLIENT_DB_HOST;
        }
    } else {
	return "$CONFIG::API{'R_CHKDBLOCK'}:0";
    }

    # if we got this far then we send the command to a remote
    &$debug("spawn remote query\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RCHKDBLOCK'};
    return &$handler_sub($target_host,$CONFIG::CLIENT_PORT);

}

# Dispatches check lock call to appropriate host
sub default_remote_check_db_lock ($$) {
    return &RatsClientLib::send_remote_command("CHKDBLOCK",@_);
}

# check the local database for a lock.
sub default_db_check_db_lock() {
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth);

    my %attr = (
                PrintError => 0,
                RaiseError => 0,
                );
    #connect to DB
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
        return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'R_CHKDBLOCK'}:1";
    } else {
	$dbh->disconnect();
	return "$CONFIG::API{'R_CHKDBLOCK'}:0";
    }
}


# checks if the db is locked. Returns 1 if locked
# 0 otherwise 
# This is used both by the API call and internally
# by various functions.
sub db_check_lock($) {
    use DBI;
    my($dbh) = $_[0];
    my ($sth, $gid, $count,$lock,$timeout,$global_lock);
    #check global lock
    $global_lock=0;
    $sth = $dbh->prepare("select count(*) from rats_locks where"
			 ." t_name = ?") || return 1;
    $sth->execute($CONFIG::USE_CLIENT_DB)|| return 1;
    $sth->bind_columns(\$global_lock);
    $sth->fetch();
    $sth->finish();
    #if locked, puke
    if($global_lock)
    {
	return 1;
    }
    return 0;
}


# search passwd file for entries 
# args: table_name,field,pattern
sub default_search(@)
{
    $main::DEBUG = 1;
    my ($where,$what,$who) = @_;
    &$debug("Got [$where,$what,$who]\n");

    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");

    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
	# if local then do it
	if ($CONFIG::CLIENT_DB_HOST eq $host) {
	    &$debug("spawn DB [$where,$what,$who]\n");
	    my $handler_sub = $CONFIG::COMMAND_HOOKS{'DSEARCH'};
	    return &$handler_sub(@_);
	# if remote then sety $target_host
	} else {
	    $target_host = $CONFIG::CLIENT_DB_HOST;
	}
    #
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
	&$debug("local [$where,$what,$who]\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'LSEARCH'};
	return &$handler_sub(@_);
    #
    } else {
	# 
	$target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }
	

    # if we got this far then we send the command to a remote
    &$debug("spawn remote [$where,$what,$who]\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RSEARCH'};
    return &$handler_sub(@_,$target_host,$CONFIG::CLIENT_PORT);
}

sub default_local_search(@)
{
    return "$CONFIG::API{'R_SEARCH'}:CALL UNSUPPORTED WITHOUT CLIENT DB"; 
}

sub default_remote_search(@)
{
    return &send_remote_command("SEARCH",@_);
}

sub default_db_search(@)
{
    my ($where,$what,$who) = @_;
    my ($sqlcrap);
    my ($uname, $retstr);
    &$debug("Got to default_db_search\n");
    use DBI;

    #set up DBI stuff
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sql, @pass_parts);
    
    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);

    # connect to database
    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});
    &$debug("connected to database\n");
    # prepare SQL statement
#    $sql = $dbh->prepare("SELECT ? FROM ? WHERE ? like ?")
#	|| return &mysql_error("Failed to prepare que);
    $sqlcrap = "SELECT $what FROM $where WHERE $what like \"$who\" AND active = \"1\"";
    $sql = $dbh->prepare($sqlcrap)
	|| return &mysql_error($CONFIG::RESPONSE{'DPF'});
    &$debug("SQL set up\n");
    # execute and fill in the blanks. Do I care about checking the return?

    $sql->execute() 
	|| return &mysql_error($CONFIG::RESPONSE{'DEF'});

    &$debug("SQL executed\n");

    #generate my return string

    my $tmpstr = $sql->fetchrow_array();
    while($uname = $sql->fetchrow_array())
    {
	$tmpstr = "$tmpstr:$uname";
    }
    $sql->finish();

    #closedb
    $dbh->disconnect();

    if($tmpstr eq "") {
	$retstr = "$CONFIG::API{'NOTFOUND'}:NO MATCHES FOUND";
    } else {
	$retstr = "$CONFIG::API{'R_SEARCH'}:$tmpstr";
    }

    return $retstr;    

}

#
# End of SEARCHPW
######

# get all password entries. returns huge string or error
sub default_get_allpwent() {

    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");
    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
	# if local then do it
	if ($CONFIG::CLIENT_DB_HOST eq $host) {
	    &$debug("spawn DB\n");
	    my $handler_sub = $CONFIG::COMMAND_HOOKS{'DGETALLPWENT'};
	    return &$handler_sub();
	# if remote then sety $target_host
	} else {
	    $target_host = $CONFIG::CLIENT_DB_HOST;
	}
    # are we doing a local file password lookup?
    } elsif ($CONFIG::ACCT_FILES_MASTER{'passwd'} eq $host) {
	&$debug("spawn local\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'LGETALLPWENT'};
	return &$handler_sub();
    # are we doing a remote lookup?
    } else {
	# figure our where the actual password file is maintained
	$target_host = $CONFIG::ACCT_FILES_MASTER{'passwd'};
    }
	

    # if we got this far then we send the command to a remote
    &$debug("spawn remote\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RGETALLPWENT'};
    return &$handler_sub($target_host,$CONFIG::CLIENT_PORT);
}


# the DB version of get_allpwent
sub default_db_get_allpwent() {
    use DBI;
    my ($return, @data);
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth, @pass_parts);

    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);

    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }

    $sth = $dbh->prepare("SELECT * FROM password_file WHERE active = \"1\"") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});
    $sth->execute() ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});

    while (@data = $sth->fetchrow_array()) {
	$data[2]--;$data[3]--;
	my $temp = join(":",@data[0 .. 6],$data[14]);
	$return .= $temp;
	$return .= "\n";

    }
    
    $sth->finish();
    $dbh->disconnect();



    if ($return) {
	return "$CONFIG::API{'R_GETALLPWENT'}:$return";
    } else {
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
}


# Dispatches getallpwent call to appropriate host
sub default_remote_get_allpwent () {
    return &RatsClientLib::send_remote_command("GETALLPWENT");
}

# get all  pwent. returns the whole passwords file or error
sub default_local_get_allpwent () {
    my ($uid, $entry, $pass_file, @passwd, $return);
    my ($usr, $pass, @rest, $target, $val);

    $val = "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    # read the passwd file, we don't care is locked
    # maybe we should?
    $pass_file = $CONFIG::ACCT_FILES{'passwd'};
    unless (open(PASSWD, $pass_file)) {
	&$log("Failed to open passwd file: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'PRE'}";
    }
    while (<PASSWD>) {
	$return .= $_;
    }
    close PASSWD;
    chomp $return;
    return "$CONFIG::API{'R_GETALLPWENT'}:$return";
}




# get all shadow entries. returns huge string or error
sub default_get_allshent() {

    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");
    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
	# if local then do it
	if ($CONFIG::CLIENT_DB_HOST eq $host) {
	    &$debug("spawn DB\n");
	    my $handler_sub = $CONFIG::COMMAND_HOOKS{'DGETALLSHENT'};
	    return &$handler_sub();
	# if remote then sety $target_host
	} else {
	    $target_host = $CONFIG::CLIENT_DB_HOST;
	}
    # are we doing a local file password lookup?
    } elsif ($CONFIG::ACCT_FILES_MASTER{'shadow'} eq $host) {
	&$debug("spawn local\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'LGETALLSHENT'};
	return &$handler_sub();
    # are we doing a remote lookup?
    } else {
	# figure our where the actual password file is maintained
	$target_host = $CONFIG::ACCT_FILES_MASTER{'shadow'};
    }
	

    # if we got this far then we send the command to a remote
    &$debug("spawn remote\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RGETALLSHENT'};
    return &$handler_sub($target_host,$CONFIG::CLIENT_PORT);
}

# the DB version of get_allshent
sub default_db_get_allshent() {
    use DBI;
    my ($return, @data,$temp);
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth, @pass_parts);
    my ($uname,$pass,$uid,$gid,$gcos,$home,$shell,$last,$min,$max,$warn);
    my ($inactive,$exp,$flag,$chg,$active);

    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);

    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }

    $sth = $dbh->prepare("SELECT * FROM password_file WHERE active = \"1\"") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});


    $sth->execute() ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});


    while (@data = $sth->fetchrow_array()) {
    ($uname,$pass,$uid,$gid,$gcos,$home,$shell,$last,$min,
     $max,$warn,$inactive,$exp,$flag,$chg,$active) = @data;

	$temp = "$uname:$pass:$last:$min:$max:$warn:$inactive:$exp:$flag:$chg";
	$return .= "\n";
	$return .= $temp

    }
    
    $sth->finish();
    $dbh->disconnect();


    if ($return) {
	return "$CONFIG::API{'R_GETALLSHENT'}:$return";
    } else {
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    }
    
}


# Dispatches getallshent call to appropriate host
sub default_remote_get_allshent () {
    return &RatsClientLib::send_remote_command("GETALLSHENT");
}

# get all  shent. returns the whole shdow file or error
sub default_local_get_allshent () {
    my ($uid, $entry, $sh_file, @shadow, $return);
    my ($usr, $pass, @rest, $target, $val);

    $val = "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSU'}";
    # read the shadow file, we don't care is locked
    # maybe we should?
    $sh_file = $CONFIG::ACCT_FILES{'shadow'};
    unless (open(SHADOW, $sh_file)) {
	&$log("Failed to open shadow file: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'SRE'}";
    }
    while (<SHADOW>) {
	$return .= $_;
    }
    close SHADOW;
    chomp $return;
    return "$CONFIG::API{'R_GETALLSHENT'}:$return";
}

# get all group entries. returns huge string or error
sub default_get_allgrent() {

    # figure out where we are
    my $host = &get_hostname();
    &$debug("determined local host as $host\n");
    my $target_host;

    # are we using a database and where?
    if ($CONFIG::USE_CLIENT_DB) {
	# if local then do it
	if ($CONFIG::CLIENT_DB_HOST eq $host) {
	    &$debug("spawn DB\n");
	    my $handler_sub = $CONFIG::COMMAND_HOOKS{'DGETALLGRENT'};
	    return &$handler_sub();
	# if remote then sety $target_host
	} else {
	    $target_host = $CONFIG::CLIENT_DB_HOST;
	}
    # are we doing a local file group lookup?
    } elsif ($CONFIG::ACCT_FILES_MASTER{'group'} eq $host) {
	&$debug("spawn local\n");
	my $handler_sub = $CONFIG::COMMAND_HOOKS{'LGETALLGRENT'};
	return &$handler_sub();
    # are we doing a remote lookup?
    } else {
	# figure our where the actual password file is maintained
	$target_host = $CONFIG::ACCT_FILES_MASTER{'group'};
    }
	

    # if we got this far then we send the command to a remote
    &$debug("spawn remote\n");
    my $handler_sub = $CONFIG::COMMAND_HOOKS{'RGETALLGRENT'};
    return &$handler_sub($target_host,$CONFIG::CLIENT_PORT);
}


# the DB version of get_allgrent
sub default_db_get_allgrent() {
    use DBI;
    my ($return, @data);
    my $dsn = "DBI:mysql:$CONFIG::USE_CLIENT_DB:localhost";
    my $db_user = "";
    my $db_pass = "";
    my ($dbh, $sth, $sth1,@members, $member, $temp, $temp2);
    my ($gname,$pass, $gid, $lastchange);

    my %attr = (
		PrintError => 0,
		RaiseError => 0,
		);

    $dbh = DBI->connect($dsn, $db_user, $db_pass, \%attr) ||
	return &mysql_error($CONFIG::RESPONSE{'DCF'});

    # check global db lock
    if (&db_check_lock($dbh)) {
	$dbh->disconnect();
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'DGL'}";
    }

    $sth = $dbh->prepare("SELECT gname,pass,gid,lastchange FROM group_file WHERE active = \"1\"") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});

    $sth1 = $dbh->prepare("SELECT uname FROM group_members WHERE gid = ?") ||
	return &mysql_error($CONFIG::RESPONSE{'DPF'});

    $sth->execute() ||
	return &mysql_error($CONFIG::RESPONSE{'DEF'});

    while (($gname, $pass, $gid, $lastchange) = $sth->fetchrow_array()) {
    #get the group members
	@members=();
	$sth1->execute($gid)||
	    return &mysql_error($CONFIG::RESPONSE{'DEF'});
	while(($member) = $sth1->fetchrow_array()) {
	    push @members, $member;
	}
	# build the response
	$temp2= join(",",@members);
	$gid--;
	$temp = "$gname:$pass:$gid:" . $temp2 . ":$lastchange";
	$return .= $temp;
	$return .= "\n";

    }



    $sth1->finish();
    $sth->finish();
    $dbh->disconnect();


    if ($return) {
	return "$CONFIG::API{'R_GETALLGRENT'}:$return";
    } else {
	return "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSG'}";
    }
    
}


# Dispatches getallgrent call to appropriate host
sub default_remote_get_allgrent () {
    return &RatsClientLib::send_remote_command("GETALLGRENT");
}

# get all group. returns the whole group file or error
sub default_local_get_allgrent () {
    my ($uid, $entry, $grp_file, @group, $return);
    my ($usr, $pass, @rest, $target, $val);

    $val = "$CONFIG::API{'NOTFOUND'}:$CONFIG::RESPONSE{'NSG'}";
    # read the group file, we don't care is locked
    # maybe we should?
    $grp_file = $CONFIG::ACCT_FILES{'group'};
    unless (open(GROUP, $grp_file)) {
	&$log("Failed to open group file: $!\n");
	return "$CONFIG::API{'CODEDERR'}:$CONFIG::RESPONSE{'GRE'}";
    }
    while (<GROUP>) {
	$return .= $_;
    }
    close GROUP;
    chomp $return;
    return "$CONFIG::API{'R_GETALLGRENT'}:$return";
}

sub get_hostname() {
    my $host = hostname();
    if ($CONFIG::FQDN_DOMAIN) {
	$host .= ".$CONFIG::FQDN_DOMAIN";
    }
    return $host;
}

# don't write anything bellow
1;
